# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging
import time

# from idlelib.debugger_r import DictProxy
from typing import Any, Dict

# Third-Party Imports
from serial import SerialException

# Local Imports
from navigate.model.devices.stage.base import StageBase
from navigate.model.devices.APIs.sutter.MP285 import MP285
from navigate.model.devices.device_types import SerialDevice
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class MP285Stage(StageBase, SerialDevice):
    """MP285Stage Class for MP-285."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        device_id: int = 0,
    ) -> None:
        """Initialize the MP285Stage.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            MP285 stage connection.
        configuration : Dict[str, Any]
            Configuration dictionary for the MP285Stage.
        device_id : int
            Device ID for the MP285Stage.

        Raises
        ------
        UserWarning
            Error while connecting to the MP285Stage.
        UserWarning
            Error while setting resolution and velocity.
        UserWarning
            Error while setting absolute operation mode.
        """
        super().__init__(microscope_name, device_connection, configuration, device_id)

        # Device Connection
        if device_connection is None:
            logger.error("The MP285 stage is unavailable.")
            raise UserWarning("The MP285 stage is unavailable.")

        #: object: MP285 stage.
        self.stage = device_connection

        #: bool: Wait until stage has finished moving before returning.
        self.stage.wait_until_done = True

        # Default mapping from self.axes to corresponding MP285 axis labelling
        axes_mapping = {"x": "x", "y": "y", "z": "z"}
        if not self.axes_mapping:
            #: dict: Dictionary of stage axes and their corresponding hardware axes.
            self.axes_mapping = {
                axis: axes_mapping[axis] for axis in self.axes if axis in axes_mapping
            }

        #: dict: Dictionary of hardware axes and their corresponding stage axes.
        self.device_axes = dict(map(lambda v: (v[1], v[0]), self.axes_mapping.items()))

        # Default Operating Parameters
        #: str: Resolution of the stage.
        self.resolution = "low"  # "high"

        #: int: Speed of the stage in units microns/s.
        self.speed = 3000

        #: float: Position of the stage along the x-axis.
        self.stage_x_pos = 0

        #: float: Position of the stage along the y-axis.
        self.stage_y_pos = 0

        #: float: Position of the stage along the z-axis.
        self.stage_z_pos = 0

        # Set the resolution and velocity of the stage
        try:
            self.stage.set_resolution_and_velocity(
                resolution=self.resolution, speed=self.speed
            )
        except Exception as e:
            logger.error(f"Communication Error: {e}")
            raise UserWarning("Sutter MP-285 - Error setting resolution and velocity")

        # Set the operating mode of the stage.
        try:
            self.stage.set_absolute_mode()
        except Exception as e:
            logger.error(f"Communication Error: {e}")
            raise UserWarning("Sutter MP-285 - Error setting absolute operation mode")

        self.report_position()

    def __del__(self) -> None:
        """Delete MP285Stage Serial Port.

        Raises
        ------
        UserWarning
            Error while closing the MP285Stage Serial Port.
        """
        self.close()

    @classmethod
    def connect(cls, port: str, baud_rate: int=115200, timeout: float=0.25) -> MP285:
        """Connect to the MP285Stage."""
        try:
            mp285_stage = MP285(port, baud_rate, timeout)
            mp285_stage.connect_to_serial()
            return mp285_stage
        except SerialException as e:
            logger.error(f"Communication Error: {e}")
            raise UserWarning(
                "Could not communicate with Sutter MP-285 via COMPORT", port
            )
        

    def report_position(self) -> dict:
        """Reports the position for all axes, and creates a position dictionary.

        Positions from the MP-285 are converted to microns.

        Returns
        -------
        position : dict
            Dictionary containing the position of all axes
        """
        position = {}
        try:
            (
                stage_x_pos,
                stage_y_pos,
                stage_z_pos,
            ) = self.stage.get_current_position()
            if stage_x_pos is not None:
                self.stage_x_pos = stage_x_pos
                self.stage_y_pos = stage_y_pos
                self.stage_z_pos = stage_z_pos
                for axis, hardware_axis in self.axes_mapping.items():
                    hardware_position = getattr(self, f"stage_{hardware_axis}_pos")
                    self.__setattr__(f"{axis}_pos", hardware_position)
            else:
                logger.debug(
                    "MP-285 didn't return current position, using previous position!"
                )

            position = self.get_position_dict()
            logger.debug(f"MP-285 - Position: {position}")
        except SerialException as e:
            print("Communication Error: {e}")
            logger.error(f"Communication Error: {e}")
            time.sleep(0.01)

        return position

    def move_axis_absolute(
        self, axis: str, abs_pos: float, wait_until_done=False
    ) -> bool:
        """Implement movement logic along a single axis.

        Parameters
        ----------
        axis : str
            An axis. For example, 'x', 'y', 'z', 'f', 'theta'.
        abs_pos : float
            Absolute position value
        wait_until_done : bool
            Block until stage has moved to its new spot.

        Returns
        -------
        bool
            Was the move successful?
        """
        move_dictionary = {f"{axis}_abs": abs_pos}
        return self.move_absolute(move_dictionary, wait_until_done)

    def move_absolute(self, move_dictionary: dict, wait_until_done=True) -> bool:
        """Move stage along a single axis.

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement. Includes 'x_abs', 'x_min',
            etc. for one or more axes. Expects values in micrometers, except for theta,
            which is in degrees.
        wait_until_done : bool
            Wait until the stage has finished moving before returning.

        Returns
        -------
        bool
            Was the move successful?
        """
        pos_dict = self.verify_abs_position(move_dictionary)
        if not pos_dict:
            return False

        # rely on cached positions
        # if len(pos_dict.keys()) < 3:
        #     self.report_position()
        self.stage.wait_until_done = wait_until_done
        move_stage = {}
        for axis in pos_dict:
            if (
                abs(
                    getattr(self, f"stage_{self.axes_mapping[axis]}_pos")
                    - pos_dict[axis]
                )
                < 0.02
            ):
                move_stage[axis] = False
            else:
                move_stage[axis] = True
                setattr(self, f"stage_{self.axes_mapping[axis]}_pos", pos_dict[axis])

        move_stage = any(move_stage.values())
        if move_stage is True:
            try:
                self.stage.move_to_specified_position(
                    x_pos=self.stage_x_pos,
                    y_pos=self.stage_y_pos,
                    z_pos=self.stage_z_pos,
                )
            except SerialException as e:
                logger.debug(f"MP285: move_axis_absolute failed - {e}")
                # make sure the cached positions are the "same" as device
                self.report_position()
                return False

        return True

    def stop(self) -> None:
        """Stop all stage movement abruptly."""
        try:
            self.stage.interrupt_move()
        except SerialException as error:
            logger.exception(f"MP-285 - Stage stop failed: {error}")

    def close(self) -> None:
        """Close the stage."""

        try:
            self.stop()
            self.stage.close()
            logger.debug("MP-285 stage connection closed")
        except (AttributeError, BaseException) as e:
            print("Error while closing the MP-285 stage connection", e)
            logger.debug("Error while disconnecting the MP-285 stage", e)



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk, simpledialog
import logging
from pathlib import Path
from typing import Optional, Callable

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.DockableNotebook import DockableNotebook
from navigate.view.custom_widgets.CollapsibleFrame import CollapsibleFrame

# Logger Setup
p = __name__.split(".")[1]

widget_types = {
    "Combobox": ttk.Combobox,
    "Input": ttk.Entry,
    "Spinbox": ttk.Spinbox,
    "Checkbutton": ttk.Checkbutton,
    "Button": ttk.Button,
}

variable_types = {
    "string": tk.StringVar,
    "float": tk.DoubleVar,
    "bool": tk.BooleanVar,
    "int": tk.IntVar,
}


class ConfigurationAssistantWindow(ttk.Frame):
    def __init__(self, root, *args, **kwargs):
        """Initiates the main application window

        Parameters
        ----------
        root : tk.Tk
            The main window of the application
        *args
            Variable length argument list
        **kwargs
            Arbitrary keyword arguments
        """
        #: tk.Tk: The main window of the application
        self.root = root
        self.root.title("Configuration Assistant")

        ttk.Frame.__init__(self, self.root, *args, **kwargs)

        #: logging.Logger: The logger for this class
        self.logger = logging.getLogger(p)

        view_directory = Path(__file__).resolve().parent
        try:
            photo_image = view_directory.joinpath("icon", "mic.png")
            self.root.iconphoto(True, tk.PhotoImage(file=photo_image))
        except tk.TclError:
            pass

        self.root.resizable(False, False)
        self.root.geometry("")
        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)

        #: ttk.Frame: The top frame of the application
        self.top_frame = ttk.Frame(self.root)

        #: ttk.Frame: The main frame of the application
        self.microscope_frame = ttk.Frame(self.root)

        self.grid(column=0, row=0, sticky=tk.NSEW)
        self.top_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=3, pady=3)
        self.microscope_frame.grid(
            row=1, column=0, columnspan=5, sticky=tk.NSEW, padx=3, pady=3
        )

        #: ttk.Frame: The top frame of the application
        self.top_window = TopWindow(self.top_frame, self.root)


class TopWindow(ttk.Frame):
    """Top Frame for Configuration Assistant.

    This class is the initial window for the configurator application.
    It contains the following:
    - Entry for number of configurations
    - Continue button
    - Cancel button
    """

    def __init__(self, main_frame, root, *args, **kwargs):
        """Initialize Top Frame.

        Parameters
        ----------
        main_frame : ttk.Frame
            Window to place widgets in.
        root : tk.Tk
            Root window of the application.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """

        #: ttk.Frame: The main frame of the application
        self.microscope_frame = main_frame
        ttk.Frame.__init__(self, self.microscope_frame, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        self.new_button = tk.Button(root, text="New Configuration")
        self.new_button.grid(row=0, column=0, sticky=tk.NE, padx=3, pady=(10, 1))
        self.new_button.config(width=15)

        self.load_button = tk.Button(root, text="Load Configuration")
        self.load_button.grid(row=0, column=1, sticky=tk.NE, padx=3, pady=(10, 1))
        self.load_button.config(width=15)

        self.add_button = tk.Button(root, text="Add A Microscope")
        self.add_button.grid(row=0, column=2, sticky=tk.NE, padx=3, pady=(10, 1))
        self.add_button.config(width=15)

        self.save_button = tk.Button(root, text="Save")
        self.save_button.grid(row=0, column=3, sticky=tk.NE, padx=3, pady=(10, 1))
        self.save_button.config(width=15)

        #: tk.Button: The button to cancel the application.
        self.cancel_button = tk.Button(root, text="Cancel")
        self.cancel_button.grid(row=0, column=4, sticky=tk.NE, padx=3, pady=(10, 1))
        self.cancel_button.config(width=15)


class MicroscopeWindow(DockableNotebook):
    def __init__(self, frame, root, *args, **kwargs):
        """Initialize Microscope Frame.

        Parameters
        ----------
        main_frame : ttk.Frame
            Window to place widgets in.
        root : tk.Tk
            Root window of the application.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """

        DockableNotebook.__init__(self, frame, root, *args, **kwargs)
        self.grid(row=0, column=0, sticky=tk.NSEW)

        self.menu.delete("Popout Tab")
        self.menu.add_command(label="Rename", command=self.rename_microscope)
        self.menu.add_command(label="Delete", command=self.delete_microscope)

    def rename_microscope(self):
        """Rename microscope"""

        if self.selected_tab_id is None:
            return
        result = simpledialog.askstring("Input", "Enter microscope name:")
        if result:
            tab_name = self.tab(self.selected_tab_id, option="text")
            self.tab(self.selected_tab_id, text=result)
            self.tab_list.remove(tab_name)
            self.tab_list.append(result)

    def delete_microscope(self):
        """Delete selected microscope"""
        if self.selected_tab_id is None:
            return
        tab_name = self.tab(self.selected_tab_id, option="text")
        self.forget(self.selected_tab_id)
        self.tab_list.remove(tab_name)


class MicroscopeTab(DockableNotebook):
    def __init__(self, parent, root, *args, **kwargs):
        """Initialize Microscope Tab.

        Parameters
        ----------
        main_frame : ttk.Frame
            Window to place widgets in.
        root : tk.Tk
            Root window of the application.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """

        # Init Frame
        DockableNotebook.__init__(self, parent, root, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

    def create_hardware_tab(
        self, name, hardware_widgets, widgets=None, top_widgets=None, **kwargs
    ):
        """Create hardware tab

        Parameters
        ----------
        name : str
            tab name/hardware name
        hardware_widgets : dict
            hardware widgets dict
        widgets : dict
            constants widgets dict
        top_widgets : dict
            button widgets dict
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments
        """
        tab = HardwareTab(
            name, hardware_widgets, widgets=widgets, top_widgets=top_widgets, **kwargs
        )
        self.tab_list.append(tab)
        self.add(tab, text=name, sticky=tk.NSEW)


class HardwareTab(ttk.Frame):
    def __init__(
        self,
        name,
        hardware_widgets,
        *args,
        widgets=None,
        top_widgets=None,
        hardware_widgets_value=[None],
        constants_widgets_value=[None],
        **kwargs
    ):
        """Initialize Microscope Tab.

        Parameters
        ----------
        name : str
            tab name/hardware name
        hardware_widgets : dict
            hardware widgets dict
        widgets : dict
            constants widgets dict
        top_widgets : dict
            button widgets dict
        hardware_widgets_value : list[dict]
           list of values for hardware widgets
        constants_widgets_value : list[dict]
           list of values for constants widgets
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments
        """
        # Init Frame
        tk.Frame.__init__(self, *args, **kwargs)

        self.name = name

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)
        scroll_frame = ttk.Frame(self)
        scroll_frame.grid(row=3, column=0, sticky=tk.NSEW)
        canvas = tk.Canvas(scroll_frame, width=1000, height=500)
        scrollbar = ttk.Scrollbar(scroll_frame, orient="vertical", command=canvas.yview)
        content_frame = ttk.Frame(canvas)

        content_frame.bind(
            "<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all"))
        )

        canvas.create_window((0, 0), window=content_frame, anchor="nw")

        canvas.configure(yscrollcommand=scrollbar.set)

        canvas.pack(side="left", fill="both", expand=True)
        scrollbar.pack(side="right", fill="y")

        self.top_frame = ttk.Frame(content_frame)

        self.top_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=10)

        self.hardware_frame = ttk.Frame(content_frame)
        self.hardware_frame.grid(row=1, column=0, sticky=tk.NSEW, padx=10)

        self.bottom_frame = ttk.Frame(content_frame)
        self.bottom_frame.grid(row=2, column=0, sticky=tk.NSEW, padx=10)
        self.frame_row = 0
        self.row_offset = self.frame_row + 1

        self.variables = {}
        self.values_dict = {}
        self.variables_list = []

        self.build_widgets(top_widgets, parent=self.top_frame)

        for widgets_value in hardware_widgets_value:
            self.build_widgets(
                hardware_widgets,
                parent=self.hardware_frame,
                widgets_value=widgets_value,
            )

        for widgets_value in constants_widgets_value:
            self.build_widgets(widgets, widgets_value=widgets_value)

    def create_hardware_widgets(self, hardware_widgets, frame, direction="vertical"):
        """create widgets

        Parameters
        ----------
        hardware_widgets : dict
            name: (display_name, widget_type, value_type, values, info)
        frame : tk.Frame
            the parent frame for widgets
        direction : str
            direction of the widget layouts
        """
        if hardware_widgets is None:
            return
        if type(frame) is CollapsibleFrame:
            content_frame = frame.content_frame
        else:
            content_frame = frame
        i = 0
        for k, v in hardware_widgets.items():
            if k == "frame_config":
                continue
            if v[1] == "Label":
                label = ttk.Label(content_frame, text=v[0])
                label.grid(row=i, column=0, sticky=tk.NW, padx=3)
                seperator = ttk.Separator(content_frame)
                seperator.grid(row=i + 1, columnspan=2, sticky=tk.NSEW, padx=3)
                i += 2
                continue
            elif v[1] != "Button":
                self.variables[k] = variable_types[v[2]]()
                label_text = v[0] + "  :" if v[0][-1] != ":" else v[0]
                label = ttk.Label(content_frame, text=label_text)
                if direction == "vertical":
                    label.grid(row=i, column=0, sticky=tk.NW, padx=(3, 10), pady=3)
                else:
                    label.grid(row=0, column=i, sticky=tk.NW, padx=(5, 3), pady=3)
                    i += 1
                if v[1] == "Checkbutton":
                    widget = widget_types[v[1]](
                        content_frame, text="", variable=self.variables[k]
                    )
                else:
                    widget = widget_types[v[1]](
                        content_frame, textvariable=self.variables[k], width=30
                    )
                if v[1] == "Combobox":
                    if type(v[3]) == list:
                        v[3] = dict([(t, t) for t in v[3]])
                    self.values_dict[k] = v[3]
                    temp = list(v[3].keys())
                    widget.config(values=temp)
                    widget.state(["!disabled", "readonly"])

                    if v[2] == "bool":
                        widget.set(str(temp[-1]))
                    else:
                        widget.set(temp[-1])
                elif v[1] == "Spinbox":
                    if type(v[3]) != dict:
                        v[3] = {}
                    widget.config(from_=v[3].get("from", 0))
                    widget.config(to=v[3].get("to", 100000))
                    widget.config(increment=v[3].get("step", 1))
                    widget.set(v[3].get("from", 0))

                # set default value
                if len(v) >= 6 and v[5] is not None:
                    self.variables[k].set(str(v[5]))
            else:
                widget = ttk.Button(
                    content_frame,
                    text=v[0],
                    command=self.build_event_handler(
                        hardware_widgets, k, frame, self.frame_row
                    ),
                )
            if direction == "vertical":
                widget.grid(row=i, column=1, sticky=tk.NSEW, padx=5, pady=3)
            else:
                widget.grid(row=0, column=i, sticky=tk.NW, padx=(10, 3), pady=(3, 0))

            # display info label
            if len(v) >= 5 and v[4]:
                label = ttk.Label(content_frame, text=v[4])
                if direction == "vertical":
                    label.grid(row=i, column=2, sticky=tk.NW, padx=(10, 10), pady=3)
                else:
                    label.grid(row=1, column=i, sticky=tk.NW, padx=(10, 3), pady=0)
            i += 1

    def build_widgets(self, widgets, *args, parent=None, widgets_value=None, **kwargs):
        """Build widgets

        Parameters
        ----------
        widgets : dict
            widget dict
        parent : frame
            parent frame to put widgets
        widgets_value : dict
            value_dict of widgets
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments, ref="reference name", direction="vertical"
        """
        if not widgets:
            return
        if parent is None:
            parent = self.bottom_frame
        collapsible = False
        title = "Hardware"
        format = None
        temp_ref = None
        direction = "vertical"
        if "frame_config" in widgets:
            collapsible = widgets["frame_config"].get("collapsible", False)
            title = widgets["frame_config"].get("title", "Hardware")
            format = widgets["frame_config"].get("format", None)
            temp_ref = widgets["frame_config"].get("ref", None)
            direction = widgets["frame_config"].get("direction", "vertical")
        if collapsible:
            self.fold_all_frames()
            frame = CollapsibleFrame(parent=parent, title=title)
            # only display one collapsible frame at a time
            frame.label.bind("<Button-1>", self.create_toggle_function(frame))
        else:
            frame = ttk.Frame(parent)
        frame.grid(row=self.frame_row, column=0, sticky=tk.NSEW, padx=20)
        self.frame_row += 1

        ref = None
        if kwargs:
            ref = kwargs.get("ref", None)
            direction = kwargs.get("direction", "vertical")
        ref = ref or temp_ref
        self.variables = {}
        self.values_dict = {}
        self.variables_list.append((self.variables, self.values_dict, ref, format))
        self.create_hardware_widgets(widgets, frame=frame, direction=direction)

        if widgets_value:
            for k, v in widgets_value.items():
                try:
                    self.variables[k].set(str(v))
                except (TypeError, ValueError):
                    pass
                except tk._tkinter.TclError:
                    pass

    def fold_all_frames(self, except_frame: Optional[tk.Frame] = None) -> None:
        """Fold all collapsible frames except one frame

        Parameters
        ----------
        except_frame : Optional[tk.Frame]
            the unfold frame
        """
        for child in self.hardware_frame.winfo_children():
            if isinstance(child, CollapsibleFrame) and child is not except_frame:
                child.fold()
        for child in self.bottom_frame.winfo_children():
            if isinstance(child, CollapsibleFrame) and child is not except_frame:
                child.fold()

    def create_toggle_function(self, frame: tk.Frame) -> Callable[..., None]:
        """Toggle collapsible frame

        Parameters
        ----------
        frame : tk.Frame
            the frame to toggle
        """

        def func(event):
            self.fold_all_frames(frame)
            frame.toggle_visibility()

        return func

    def build_event_handler(
        self, hardware_widgets: dict, key: str, frame: tk.Frame, frame_id: int
    ) -> Callable[..., None]:
        """Build button event handler

        Parameters
        ----------
        hardware_widgets : dict
            widget dict containing the button
        key : str
            reference of the button
        frame : tk.Frame
            the frame to put/delete widgets
        frame_id : int
            index of the frame

        Returns
        -------
        func : Callable
            event handler function
        """

        def func(*args, **kwargs):
            v = hardware_widgets[key]
            if "widgets" in v[2]:
                if "parent" in v[2]:
                    parent = (
                        self.hardware_frame
                        if v[2]["parent"].startswith("hardware")
                        else None
                    )
                else:
                    parent_id = frame.winfo_parent()
                    parent = self.nametowidget(parent_id)
                widgets = (
                    hardware_widgets if v[2]["widgets"] == "self" else v[2]["widgets"]
                )
                self.build_widgets(
                    widgets,
                    parent=parent,
                    ref=v[2].get("ref", None),
                    direction=v[2].get("direction", "vertical"),
                )
                # collapse other frame
            elif v[2].get("delete", False):
                frame.grid_remove()
                self.variables_list[frame_id - self.row_offset] = None

        return func



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
import logging
from typing import Any, Dict

# Third Party Imports

# Local Imports
from navigate.model.waveforms import (
    remote_focus_ramp,
    smooth_waveform,
    remote_focus_ramp_triangular,
)
from navigate.tools.decorators import log_initialization

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class RemoteFocusBase:
    """RemoteFocusBase Class - Parent class for Remote Focusing Device."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args,
        **kwargs,
    ) -> None:
        """Initializes the RemoteFocusBase Class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            Device connection object.
        configuration : Dict[str, Any]
            Configuration dictionary.
        """

        #: Any: Device connection object.
        self.device_connection = device_connection

        #: dict: Configuration dictionary.
        self.configuration = configuration

        #: str: Name of the microscope.
        self.microscope_name = microscope_name

        #: dict: Remote focus device parameters.
        self.device_config = configuration["configuration"]["microscopes"][
            microscope_name
        ]["remote_focus"]

        #: int: Sample rate of the DAQ.
        self.sample_rate = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["sample_rate"]

        #: float: Sweep time of the DAQ.
        self.sweep_time = 0

        #: float: Camera delay percent.
        self.camera_delay = (
            configuration["configuration"]["microscopes"][microscope_name]["camera"][
                "delay"
            ]
            / 1000
        )

        # Waveform Parameters
        #: float: Remote focus max voltage.
        self.remote_focus_max_voltage = self.device_config["hardware"]["max"]

        #: float: Remote focus min voltage.
        self.remote_focus_min_voltage = self.device_config["hardware"]["min"]

        #: dict: Waveform dictionary.
        self.waveform_dict = {}

    def __str__(self):
        """String representation of the RemoteFocusBase class."""
        return "RemoteFocusBase"

    def __del__(self):
        """Destructor"""
        pass

    def adjust(self, exposure_times, sweep_times, offset=None):
        """Adjusts the remote focus waveform based on the readout time.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel
        offset : float, optional
            Offset value for the remote focus waveform, by default None

        Returns
        -------
        waveform : numpy.ndarray
            Waveform for the remote focus device.
        """
        # to determine if the waveform has to be triangular
        sensor_mode = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["sensor_mode"]
        readout_direction = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["readout_direction"]

        self.waveform_dict = dict.fromkeys(self.waveform_dict, None)
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        waveform_constants = self.configuration["waveform_constants"]
        imaging_mode = microscope_state["microscope_name"]
        zoom = microscope_state["zoom"]
        # ramp_type = self.configuration["configuration"]["microscopes"][
        #     self.microscope_name]['remote focus device']['ramp_type']
        self.sample_rate = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["daq"]["sample_rate"]

        remote_focus_ramp_falling = (
            float(waveform_constants["other_constants"]["remote_focus_ramp_falling"])
            / 1000
        )
        remote_focus_delay = (
            float(waveform_constants["other_constants"]["remote_focus_delay"]) / 1000
        )
        percent_smoothing = float(
            waveform_constants["other_constants"]["percent_smoothing"]
        )

        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                # Get the Waveform Parameters - Assumes ETL Delay < Camera Delay.
                # Should Assert.
                laser = channel["laser"]
                exposure_time = exposure_times[channel_key]
                self.sweep_time = sweep_times[channel_key]

                samples = int(self.sample_rate * self.sweep_time)

                # Remote Focus Parameters
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["amplitude"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"] = "0"

                remote_focus_amplitude = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"]
                )

                # Validation for when user puts a '-' in spinbox
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["offset"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"] = "0"

                remote_focus_offset = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"]
                )
                if offset is not None:
                    remote_focus_offset += offset

                # Calculate the Waveforms
                if sensor_mode == "Light-Sheet" and (
                    readout_direction == "Bidirectional"
                    or readout_direction == "Rev. Bidirectional"
                ):
                    self.waveform_dict[channel_key] = remote_focus_ramp_triangular(
                        sample_rate=self.sample_rate,
                        exposure_time=exposure_time,
                        sweep_time=self.sweep_time,
                        remote_focus_delay=remote_focus_delay,
                        camera_delay=self.camera_delay,
                        amplitude=remote_focus_amplitude,
                        offset=remote_focus_offset,
                    )
                    samples *= 2

                else:
                    self.waveform_dict[channel_key] = remote_focus_ramp(
                        sample_rate=self.sample_rate,
                        exposure_time=exposure_time,
                        sweep_time=self.sweep_time,
                        remote_focus_delay=remote_focus_delay,
                        camera_delay=self.camera_delay,
                        fall=remote_focus_ramp_falling,
                        amplitude=remote_focus_amplitude,
                        offset=remote_focus_offset,
                    )

                # Smooth the Waveform if specified
                if percent_smoothing > 0:
                    self.waveform_dict[channel_key] = smooth_waveform(
                        waveform=self.waveform_dict[channel_key],
                        percent_smoothing=percent_smoothing,
                    )[:samples]

                # Clip any values outside the hardware limits
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] > self.remote_focus_max_voltage
                ] = self.remote_focus_max_voltage
                self.waveform_dict[channel_key][
                    self.waveform_dict[channel_key] < self.remote_focus_min_voltage
                ] = self.remote_focus_min_voltage

        return self.waveform_dict



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library imports
import logging
import importlib  # noqa: F401
from multiprocessing.managers import ListProxy
import reprlib
from typing import Any, Dict, List, Optional

# Third-party imports
import numpy as np

# Local application imports
from navigate.model.device_startup_functions import load_devices, start_device
from navigate.tools.common_functions import build_ref_name

# Set up logging
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Microscope:
    """Microscope Class - Used to control the microscope."""

    def __init__(
        self,
        name: str,
        configuration: Dict[str, Any],
        devices_dict: dict,
        is_synthetic: bool = False,
        is_virtual: bool = False,
    ) -> None:
        """Initialize the microscope.

        Parameters
        ----------
        name : str
            Name of the microscope.
        configuration : Dict[str, Any]
            Configuration dictionary.
        devices_dict : dict
            Dictionary of devices.
        is_synthetic : bool, optional
            Is synthetic, by default False
        is_virtual : bool, optional
            Is virtual, by default False
        """

        # Initialize microscope object
        #: str: Name of the microscope.
        self.microscope_name = name

        #: Queue: Output event queue.
        self.output_event_queue = None

        #: dict: Configuration dictionary.
        self.configuration = configuration

        #: SharedNDArray: Buffer for image data.
        self.data_buffer = None

        #: dict: Dictionary of stages.
        self.stages = {}

        #: list: List of stages.
        self.stages_list = []

        #: bool: Ask stage for position.
        self.ask_stage_for_position = True

        #: obj: Camera object.
        self.camera = None

        #: Any: Shutter device.
        self.shutter = None

        #: dict: Dictionary of lasers.
        self.laser = {}

        #: dict: Dictionary of galvanometers.
        self.galvo = {}

        #: Any: Remote focus device.
        self.remote_focus = None

        #: dict: Dictionary of filter_wheels
        self.filter_wheel = {}

        #: dict: Dictionary of microscope info.
        self.info = {}

        #: int: Current channel.
        self.current_channel = None

        #: int: Current laser index.
        self.current_laser_index = 0

        #: list: List of all channels.
        self.channels = None

        #: list: List of available channels.
        self.available_channels = None

        #: int: Number of images.
        self.number_of_frames = None

        #: float: Central focus position.
        self.central_focus = None

        #: Bool: Is a synthetic microscope.
        self.is_synthetic = is_synthetic

        #: list: List of laser wavelengths.
        self.laser_wavelength = []

        #: dict: Dictionary of returned stage positions.
        self.ret_pos_dict = {}

        #: dict: Dictionary of commands
        self.commands = {}

        #: dict: Dictionary of plugin devices
        self.plugin_devices = {}

        if is_virtual:
            return

        device_ref_dict = {
            "camera": ["serial_number"],
            "filter_wheel": ["type", "wheel_number"],
            "zoom": ["type", "servo_id"],
            "shutter": ["type", "channel"],
            "remote_focus": ["type", "channel"],
            "galvo": ["type", "channel"],
            "laser": ["wavelength"],
            "mirror": ["type"],
        }

        device_name_dict = {"laser": "wavelength"}

        laser_list = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["laser"]

        #: dict: Dictionary of laser configurations.
        self.laser_wavelength = [laser["wavelength"] for laser in laser_list]

        if "__plugins__" not in devices_dict:
            devices_dict["__plugins__"] = {}

        devices_dict = load_devices(
            self.microscope_name,
            configuration,
            is_synthetic,
            devices_dict,
            devices_dict["__plugins__"],
        )
        daq_type = (
            "Synthetic"
            if is_synthetic
            else self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ]["daq"]["hardware"]["type"]
        )
        #: dict: Dictionary of data acquisition devices.
        self.daq = devices_dict["daq"].get(daq_type, None)

        # Load and start all devices
        for device_name in self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ].keys():

            # Skip the daq and stage devices. These are handled separately.
            if device_name in ["daq", "stage"]:
                continue
            is_plugin = False
            device_connection = None
            (
                device_config_list,
                device_name_list,
                is_list,
            ) = self.assemble_device_config_lists(
                device_name=device_name, device_name_dict=device_name_dict
            )
            # new device type
            if device_name not in device_ref_dict:
                if device_name in devices_dict["__plugins__"]:
                    device_ref_dict[device_name] = devices_dict["__plugins__"][
                        device_name
                    ]["ref_list"]
                    is_plugin = True
                else:
                    print(
                        f"Device {device_name} could not be loaded! "
                        f"Please make sure there is no spelling error!"
                    )
                    logger.debug(
                        f"Device {device_name} could not be loaded! "
                        f"Please make sure there is no spelling error!"
                    )
                    continue
            for i, device in enumerate(device_config_list):
                device_ref_name = None
                try:
                    ref_list = [
                        device["hardware"][k] for k in device_ref_dict[device_name]
                    ]
                except Exception as e:
                    logger.error(
                        f"Can't get the device attributes in configuration file: {e}"
                    )

                device_ref_name = build_ref_name("_", *ref_list)
                if (
                    device_name in devices_dict
                    and device_ref_name in devices_dict[device_name]
                ):
                    # get the device
                    device_connection = devices_dict[device_name][device_ref_name]

                if is_plugin:
                    device_plugin_dict = devices_dict.get(device_name, {})
                    try:
                        if device_connection is None:
                            exec(
                                f"device_plugin_dict['{device_ref_name}'] = devices_dict["
                                f"'__plugins__']['{device_name}']['load_device']"
                                f"(configuration['configuration']['microscopes']["
                                f"self.microscope_name]['{device_name}']['hardware'], "
                                f"is_synthetic)"
                            )
                            devices_dict[device_name] = device_plugin_dict
                            device_connection = device_plugin_dict[device_ref_name]
                        exec(
                            f"self.plugin_devices['{device_name}'] = devices_dict["
                            f"'__plugins__']['{device_name}']['start_device']"
                            f"(self.microscope_name, device_connection, configuration, "
                            f"is_synthetic)"
                        )
                    except RuntimeError:
                        print(
                            f"Device {device_name} isn't loaded correctly! "
                            f"Please check the spelling and the plugin!"
                        )
                        continue

                    self.info[device_name] = device_ref_name
                    commands_dict = self.plugin_devices[device_name].commands
                    for command in commands_dict:
                        self.commands[command] = (device_name, commands_dict[command])
                    continue

                # LOAD AND START DEVICES
                self.load_and_start_devices(
                    device_name=device_name,
                    is_list=is_list,
                    device_name_list=device_name_list,
                    device_ref_name=device_ref_name,
                    device_connection=device_connection,
                    name=name,
                    i=i,
                    plugin_devices=devices_dict["__plugins__"],
                )

                if device_connection is None and device_ref_name is not None:
                    if device_name not in devices_dict:
                        devices_dict[device_name] = {}

                    devices_dict[device_name][device_ref_name] = (
                        getattr(self, device_name)[device_name_list[i]]
                        if is_list
                        else getattr(self, device_name)
                    )

        # stages
        stage_devices = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["stage"]["hardware"]

        # set the NI Galvo stage flag
        self.configuration["configuration"]["microscopes"][self.microscope_name][
            "stage"
        ]["has_ni_galvo_stage"] = False
        if type(stage_devices) != ListProxy:
            stage_devices = [stage_devices]

        for i, device_config in enumerate(stage_devices):
            device_ref_name = build_ref_name(
                "_", device_config["type"], device_config["serial_number"]
            )
            if device_config["type"] == "NIStage":
                self.configuration["configuration"]["microscopes"][
                    self.microscope_name
                ]["stage"]["has_ni_galvo_stage"] = True

            try:
                stage = start_device(
                    microscope_name=self.microscope_name,
                    configuration=self.configuration,
                    device_category="stage",
                    device_id=i,
                    is_synthetic=is_synthetic,
                    daq_connection=self.daq,
                    plugin_devices=devices_dict["__plugins__"],
                )
            except Exception as e:
                raise Exception(
                    "Stage not found. "
                    "This often arises when the configuration.yaml file is "
                    "incorrectly specified. Please check the "
                    "configuration.yaml file and try again. Things to "
                    "check include: ",
                    "1. If no stage is physically connected to the "
                    "microscope, one must still be listed in the "
                    "configuration.yaml file. We recommend that you use a"
                    "`SyntheticStage`. "
                    "2. If a stage is physically connected to the "
                    "microscope, make sure that it is corrected defined in "
                    "both the `hardware` and `microscopes` sections of the "
                    "configuration.yaml file. Importantly, name, "
                    "type, and serial numbers must match. "
                    "3. You may be using a stage that is not supported by "
                    "Navigate. Please check the list of supported stages in "
                    "the documentation."
                    f"The stage that failed to load is: {device_ref_name}",
                )
            for axis in device_config["axes"]:
                self.stages[axis] = stage
                self.info[f"stage_{axis}"] = device_ref_name

            self.stages_list.append((stage, list(device_config["axes"])))

        # connect daq and camera in synthetic mode
        if is_synthetic and self.daq is not None:
            self.daq.add_camera(self.microscope_name, self.camera)

    def update_data_buffer(
        self, data_buffer: List[np.ndarray], number_of_frames: int
    ) -> None:
        """Update the data buffer for the camera.

        Parameters
        ----------
        data_buffer : List[np.ndarray]
            Data buffer for the camera.
        number_of_frames : int
            Number of frames to be acquired.
        """

        if self.camera.is_acquiring:
            self.camera.close_image_series()
        self.data_buffer = data_buffer
        self.number_of_frames = number_of_frames

    def move_stage_offset(self, former_microscope: Optional[str] = None) -> None:
        """Move the stage to the offset position.

        Parameters
        ----------
        former_microscope : Optional[str], optional
            Name of the former microscope.
        """

        if former_microscope:
            former_offset_dict = self.configuration["configuration"]["microscopes"][
                former_microscope
            ]["stage"]
        else:
            former_offset_dict = dict((f"{a}_offset", 0) for a in self.stages)
        self_offset_dict = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["stage"]
        self.ask_stage_for_position = True
        pos_dict = self.get_stage_position()
        for stage, axes in self.stages_list:

            # x_abs: current x_pos + current_x_offset - former_x_offset
            pos = {
                axis
                + "_abs": (
                    pos_dict[axis + "_pos"]
                    + self_offset_dict.get(axis + "_offset", 0)
                    - former_offset_dict.get(axis + "_offset", 0)
                )
                for axis in axes
            }
            stage.move_absolute(pos, wait_until_done=True)
        self.ask_stage_for_position = True

    def prepare_acquisition(self) -> dict:
        """Prepare the acquisition.

        This function prepares the acquisition by identifying which channels are
        selected for imaging, setting the camera region of interest (ROI), setting
        the camera sensor mode, setting the camera binning, initializing the image
        series, calculating all the waveforms required for the acquisition process,
        and opens the shutter.

        Returns
        -------
        waveform : dict/None
            Dictionary of all the waveforms. None if the camera ROI and binning are
            not set successfully.
        """
        self.current_channel = 0
        self.central_focus = None
        self.get_available_channels()

        if self.camera.is_acquiring:
            self.camera.close_image_series()
        # set camera trigger source
        self.set_camera_trigger_mode()
        self.set_camera_sensor_mode()
        if not self.set_camera_roi_and_binning():
            return None
        logger.debug(f"Running microscope {self.microscope_name}")
        self.report_camera_settings()
        # Initialize Image Series - Attaches camera buffer and start imaging
        self.camera.initialize_image_series(self.data_buffer, self.number_of_frames)

        # calculate all the waveform
        self.shutter.open_shutter()

        return self.calculate_all_waveform()

    def get_available_channels(self) -> None:
        """Get the available channels for imaging.

        This function gets the available channels for imaging by identifying which
        channels are selected for imaging in the configuration file.
        """
        self.channels = self.configuration["experiment"]["MicroscopeState"]["channels"]
        self.available_channels = list(
            map(
                lambda c: int(c[len("channel_") :]),
                filter(lambda k: self.channels[k]["is_selected"], self.channels.keys()),
            )
        )

    def report_camera_settings(self) -> None:
        """Log the camera settings.

        This function logs the camera settings for the current acquisition. It logs the
        camera parameters specified in the configuration file for the current
        microscope.
        """
        camera_info = reprlib.Repr()
        camera_info.indent = "---"
        camera_info.maxdict = 20
        camera_info = camera_info.repr(
            dict(
                self.configuration["experiment"]["CameraParameters"][
                    self.microscope_name
                ]
            )
        )
        logger.info(f"Preparing Acquisition. Camera Parameters: {camera_info}")

    def set_camera_trigger_mode(self) -> None:
        """Set the camera trigger mode.

        This function sets the camera trigger source, trigger mode."""
        trigger_source = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ].get("trigger_source", "External")
        self.camera.set_trigger_mode(trigger_source)

    def set_camera_sensor_mode(self) -> None:
        """Set the camera sensor mode.

        This function sets the camera sensor mode based on the camera parameters
        specified in the configuration file. It sets the sensor mode and the readout
        direction for the camera if it is in the light-sheet imaging mode.
        """
        # Set Camera Sensor Mode - Must be done before camera is initialized.
        sensor_mode = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["sensor_mode"]
        self.camera.set_sensor_mode(sensor_mode)
        if sensor_mode == "Light-Sheet":
            self.camera.set_readout_direction(
                self.configuration["experiment"]["CameraParameters"][
                    self.microscope_name
                ]["readout_direction"]
            )

    def set_camera_roi_and_binning(self) -> bool:
        """Set the camera ROI and binning.

        This function sets the camera region of interest (ROI) based on the camera
        parameters specified in the configuration file. It sets the image width,
        image height, and the center of the image. The camera ROI is used to define
        the area of the image sensor that will be used to capture images during the
        acquisition process. The function also sets the camera binning value.

        Returns
        -------
        bool
            True if the camera ROI and binning are set successfully, False otherwise.
        """

        img_width = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["x_pixels"]
        img_height = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["y_pixels"]
        center_x = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["center_x"]
        center_y = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["center_y"]
        self.camera.set_ROI(img_width, img_height, center_x, center_y)
        return self.camera.set_ROI_and_binning(
            img_width,
            img_height,
            center_x,
            center_y,
            self.configuration["experiment"]["CameraParameters"][self.microscope_name][
                "binning"
            ],
        )

    def end_acquisition(self) -> None:
        """End the acquisition.

        This function stops the acquisition, which includes stopping the data
        acquisition system, closing the camera image series, closing the shutter,
        turning off the lasers, and moving the stage to the central focus
        position if it was moved during the acquisition. It also stops the stage if
        it is moving.

        For ASI daq systems, it also stops the ASI galvos (TG-1000 waveforms in SAM mode
        4 must be manually shut off).
        """
        self.daq.stop_acquisition()
        galvo_type = self.configuration["configuration"]["microscopes"][
            self.microscope_name
        ]["galvo"][0]["hardware"]["type"]
        if galvo_type == "asi.ASI":
            for k in self.galvo:
                self.galvo[k].turn_off()

        self.stop_stage()
        if self.central_focus is not None:
            self.move_stage({"f_abs": self.central_focus})
        if self.camera.is_acquiring:
            self.camera.close_image_series()
        self.shutter.close_shutter()
        for k in self.laser:
            self.laser[k].turn_off()
        self.current_channel = 0
        self.central_focus = None
        logger.info("Acquisition Ended")

    def turn_on_laser(self) -> None:
        """Turn on the current laser."""
        logger.info(
            f"Turning on laser {self.laser_wavelength[self.current_laser_index]}"
        )
        self.laser[str(self.laser_wavelength[self.current_laser_index])].turn_on()

    def turn_off_lasers(self) -> None:
        """Turn off current laser."""
        logger.info(
            f"Turning off laser {self.laser_wavelength[self.current_laser_index]}"
        )
        self.laser[str(self.laser_wavelength[self.current_laser_index])].turn_off()

    def calculate_all_waveform(self) -> dict:
        """Calculate all the waveforms.

        Returns
        -------
        waveform : dict
            Dictionary of all the waveforms.
        """
        exposure_times, sweep_times = self.calculate_exposure_sweep_times()
        camera_waveform = self.daq.calculate_all_waveforms(
            self.microscope_name, exposure_times, sweep_times
        )
        remote_focus_waveform = self.remote_focus.adjust(exposure_times, sweep_times)
        galvo_waveform = [
            self.galvo[k].adjust(exposure_times, sweep_times) for k in self.galvo
        ]

        # calculate waveform for galvo stage
        for stage, _ in self.stages_list:
            if type(stage).__name__ == "GalvoNIStage":
                stage.switch_mode("normal", exposure_times, sweep_times)
        waveform_dict = {
            "camera_waveform": camera_waveform,
            "remote_focus_waveform": remote_focus_waveform,
            "galvo_waveform": galvo_waveform,
        }
        return waveform_dict

    def calculate_exposure_sweep_times(self) -> tuple:
        """Calculate the exposure and sweep times for all channels.

        The `calculate_exposure_sweep_times` function calculates and returns exposure
        times and sweep times for all channels in a microscope configuration. It takes
        the camera's readout time as an input parameter and considers various parameters
        such as camera exposure time, delay percentages, and smoothing to compute these
        times. The function iterates through the channels, performs calculations, and
        returns the results as dictionaries containing exposure times and sweep times
        for each channel.

        Returns
        -------
        exposure_times : dict
            Dictionary of exposure times.
        sweep_times : dict
            Dictionary of sweep times.
        """
        exposure_times = {}
        sweep_times = {}
        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        waveform_constants = self.configuration["waveform_constants"]

        logger.info(f"Microscope state: {repr(dict(microscope_state))}")
        logger.info(f"Waveform constants: {repr(dict(waveform_constants))}")

        camera_delay = (
            self.configuration["configuration"]["microscopes"][self.microscope_name][
                "camera"
            ]["delay"]
            / 1000
        )
        camera_settle_duration = (
            self.configuration["configuration"]["microscopes"][self.microscope_name][
                "camera"
            ].get("settle_duration", 0)
            / 1000
        )
        remote_focus_ramp_falling = (
            float(waveform_constants["other_constants"]["remote_focus_ramp_falling"])
            / 1000
        )

        duty_cycle_wait_duration = (
            float(waveform_constants["other_constants"]["remote_focus_settle_duration"])
            / 1000
        )
        ps = float(waveform_constants["other_constants"].get("percent_smoothing", 0.0))

        readout_time = 0
        readout_mode = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["sensor_mode"]

        if readout_mode == "Normal":
            readout_time = self.camera.calculate_readout_time()
        elif self.configuration["experiment"]["CameraParameters"][self.microscope_name][
            "readout_direction"
        ] in ["Bidirectional", "Rev. Bidirectional"]:
            remote_focus_ramp_falling = 0
        # set readout out time
        self.configuration["experiment"]["CameraParameters"][self.microscope_name][
            "readout_time"
        ] = (readout_time * 1000)

        for channel_key in microscope_state["channels"].keys():
            channel = microscope_state["channels"][channel_key]
            if channel["is_selected"] is True:
                exposure_time = float(channel["camera_exposure_time"]) / 1000

                if readout_mode == "Light-Sheet":
                    (
                        _,
                        _,
                        updated_exposure_time,
                    ) = self.camera.calculate_light_sheet_exposure_time(
                        exposure_time,
                        int(
                            self.configuration["experiment"]["CameraParameters"][
                                self.microscope_name
                            ]["number_of_pixels"]
                        ),
                    )
                    if updated_exposure_time != exposure_time:
                        print(
                            f"*** Notice: The actual exposure time of the camera for "
                            f"{channel_key} is {round(updated_exposure_time*1000, 1)}"
                            f"ms, not {exposure_time*1000}ms!"
                        )
                        exposure_time = round(updated_exposure_time, 4)
                        # update the experiment file
                        channel["camera_exposure_time"] = round(
                            updated_exposure_time * 1000, 1
                        )
                        self.output_event_queue.put(
                            (
                                "exposure_time",
                                (channel_key, channel["camera_exposure_time"]),
                            )
                        )

                sweep_time = (
                    exposure_time
                    + readout_time
                    + camera_delay
                    + max(
                        remote_focus_ramp_falling + duty_cycle_wait_duration,
                        camera_settle_duration,
                        camera_delay,
                    )
                    - camera_delay
                )
                # TODO: should we keep the percent_smoothing?
                if ps > 0:
                    sweep_time = (1 + ps / 100) * sweep_time

                exposure_times[channel_key] = exposure_time + readout_time
                sweep_times[channel_key] = sweep_time

        self.exposure_times = exposure_times
        self.sweep_times = sweep_times

        return exposure_times, sweep_times

    def get_exposure_sweep_times(self) -> tuple:
        """Get the exposure and sweep times for all channels.

        Returns
        -------
        exposure_times : dict
            Dictionary of exposure times.
        sweep_times : dict
            Dictionary of sweep times.

        """
        return self.exposure_times, self.sweep_times

    def prepare_next_channel(self, update_daq_task_flag: bool = True) -> None:
        """Prepare the next channel.

        This function, `prepare_next_channel`, is responsible for configuring various
        hardware components for the next imaging channel in an experimental setup.
        It sequentially selects the next available channel, sets the filter wheel,
        camera exposure time, laser power, and other parameters based on the selected
        channel's configuration. Additionally, it stops data acquisition, prepares the
        data acquisition system for the new channel, and adjusts the focus position as
        necessary, ensuring the hardware is ready for imaging the selected channel.

        Parameters
        ----------
        update_daq_task_flag : bool
            whether to override waveforms in the DAQ (create new tasks)
        """
        curr_channel = self.current_channel
        if self.current_channel == 0:
            self.current_channel = self.available_channels[0]
        else:
            idx = (self.available_channels.index(self.current_channel) + 1) % len(
                self.available_channels
            )
            self.current_channel = self.available_channels[idx]
        if curr_channel == self.current_channel:
            return

        prefix = "channel_"
        channel_key = prefix + str(self.current_channel)
        channel = self.configuration["experiment"]["MicroscopeState"]["channels"][
            channel_key
        ]
        # Filter Wheel Settings.
        for k in self.filter_wheel:
            self.filter_wheel[k].set_filter(channel[k])

        # Camera Settings
        self.set_camera_exposure_time(channel)

        # Laser Settings
        self.current_laser_index = channel["laser_index"]
        for k in self.laser:
            self.laser[k].turn_off()
        self.laser[str(self.laser_wavelength[self.current_laser_index])].set_power(
            channel["laser_power"]
        )
        logger.info(
            f"{self.laser_wavelength[self.current_laser_index]} "
            f"nm laser power set to {channel['laser_power']}"
        )
        # self.laser[str(self.laser_wavelength[self.current_laser_index])].turn_on()

        # stop daq before writing new waveform
        # When called the first time, throws an error.
        # choose to not update the waveform is very useful when running ZStack
        # if there is a NI Galvo stage in the system.
        if update_daq_task_flag:
            self.daq.stop_acquisition()
            self.daq.prepare_acquisition(channel_key)

        # Add Defocus term
        # Assume wherever we start is the central focus
        # TODO: is this the correct assumption?
        if self.central_focus is None:
            self.central_focus = self.get_stage_position().get("f_pos")
        if self.central_focus is not None:
            self.move_stage(
                {"f_abs": self.central_focus + float(channel["defocus"])},
                wait_until_done=True,
                update_focus=False,
            )

    def set_camera_exposure_time(self, channel: dict) -> None:
        """Set the camera exposure time.

        This function prepares the camera for imaging by setting the camera exposure
        time based on the selected channel's configuration. It also adjusts the
        camera line interval if the sensor mode is set to the Light-Sheet mode.

        Parameters
        ----------
        channel : dict
            Dictionary of channel parameters.
        """
        self.current_exposure_time = float(channel["camera_exposure_time"]) / 1000
        if (
            self.configuration["experiment"]["CameraParameters"][self.microscope_name][
                "sensor_mode"
            ]
            == "Light-Sheet"
        ):
            (
                self.current_exposure_time,
                camera_line_interval,
                _,
            ) = self.camera.calculate_light_sheet_exposure_time(
                self.current_exposure_time,
                int(
                    self.configuration["experiment"]["CameraParameters"][
                        self.microscope_name
                    ]["number_of_pixels"]
                ),
            )
            self.camera.set_line_interval(camera_line_interval)
            logger.info(f"Camera line interval set to {camera_line_interval}.")
        self.camera.set_exposure_time(self.current_exposure_time)
        logger.info(f"Camera exposure time set to {self.current_exposure_time}.")

    def move_stage(
        self, pos_dict: dict, wait_until_done: bool = False, update_focus: bool = True
    ) -> bool:
        """Move stage to a position.

        Parameters
        ----------
        pos_dict : dict
            Dictionary of stage positions.
        wait_until_done : bool, optional
            Wait until stage is done moving, by default False
        update_focus : bool, optional
            Update the central focus

        Returns
        -------
        success : bool
            True if stage is successfully moved, False otherwise.
        """
        self.ask_stage_for_position = True

        if len(pos_dict.keys()) == 1:
            axis_key = list(pos_dict.keys())[0]
            axis = axis_key[: axis_key.index("_")]
            if update_focus and axis == "f":
                self.central_focus = None
            return self.stages[axis].move_axis_absolute(
                axis, pos_dict[axis_key], wait_until_done
            )

        success = True
        for stage, axes in self.stages_list:
            pos = {
                axis: pos_dict[axis]
                for axis in pos_dict
                if axis[: axis.index("_")] in axes
            }
            if pos:
                success = stage.move_absolute(pos, wait_until_done) and success

        if update_focus and "f_abs" in pos_dict:
            self.central_focus = None

        return success

    def stop_stage(self) -> None:
        """Stop stage."""

        self.ask_stage_for_position = True

        for stage, axes in self.stages_list:
            stage.stop()

        self.central_focus = self.get_stage_position().get("f_pos", self.central_focus)

    def get_stage_position(self) -> dict:
        """Get stage position.

        Returns
        -------
        stage_position : dict
            Dictionary of stage positions.
        """
        if self.ask_stage_for_position:
            for stage, axes in self.stages_list:
                temp_pos = stage.report_position()
                self.ret_pos_dict.update(temp_pos)
            self.ask_stage_for_position = False
        return self.ret_pos_dict

    def move_remote_focus(self, offset: Optional[float] = None) -> None:
        """Move remote focus.

        Parameters
        ----------
        offset : Optional[float], optional
            Offset, by default None
        """
        exposure_times, sweep_times = self.calculate_exposure_sweep_times()
        self.remote_focus.move(exposure_times, sweep_times, offset)

    def update_stage_limits(self, limits_flag: bool = True) -> None:
        """Update stage limits.

        Parameters
        ----------
        limits_flag : bool, optional
            Limits flag, by default True
        """
        logger.info(f"Stage limits enabled: {limits_flag}")
        self.ask_stage_for_position = True
        for stage, _ in self.stages_list:
            stage.stage_limits = limits_flag

    def assemble_device_config_lists(
        self, device_name: str, device_name_dict: dict
    ) -> tuple:
        """Assemble device config lists.

        Parameters
        ----------
        device_name : str
            Device name.
        device_name_dict : dict
            Device name dictionary.

        Returns
        -------
        device_config_list : list
            Device configuration list.
        device_name_list : list
            Device name list.
        is_list : bool
            Is list.
        """
        device_config_list = []
        device_name_list = []

        try:
            devices = self.configuration["configuration"]["microscopes"][
                self.microscope_name
            ][device_name]
        except KeyError:
            # if no such device
            return [], [], False

        if type(devices) == ListProxy:
            i = 0
            for d in devices:
                device_config_list.append(d)
                if device_name in device_name_dict:
                    device_name_list.append(
                        build_ref_name("_", d[device_name_dict[device_name]])
                    )

                else:
                    device_name_list.append(build_ref_name("_", device_name, i))
                i += 1
            is_list = True

        else:
            device_config_list.append(devices)
            is_list = False

        return device_config_list, device_name_list, is_list

    def load_and_start_devices(
        self,
        device_name: str,
        is_list: bool,
        device_name_list: list,
        device_ref_name: str,
        device_connection: Any,
        name: str,
        i: int,
        plugin_devices: dict,
    ) -> None:
        """Load and start devices.

        Function uses importlib to import the device startup functions and then
        starts the devices. If it is a list, it will start the device at the index i.
        If it is a plugin, it will load the plugin device.

        Parameters
        ----------
        device_name : str
            The name of the device.
        is_list : bool
            Whether the device is a list of devices.
        device_name_list : list
            The list of device names.
        device_ref_name : str
            The reference name of the device.
        device_connection : Any
            The communication instance of the device.
        name : str
            The name of the microscope.
        i : int
            Index.
        plugin_devices : dict
            Plugin Devices

        """

        # Start the devices
        if is_list:
            exec(
                f"self.{device_name}['{device_name_list[i]}'] = "
                f"start_device(name, self.configuration, '{device_name}', "
                f"i, self.is_synthetic, self.daq, plugin_devices)"
            )
            if device_name in device_name_list[i]:
                self.info[device_name_list[i]] = device_ref_name
        else:
            exec(
                f"self.{device_name} = start_device(name, "
                f"self.configuration, '{device_name}', 0, "
                f"self.is_synthetic, self.daq, plugin_devices)"
            )
            self.info[device_name] = device_ref_name

    def terminate(self) -> None:
        """Close hardware explicitly.

        Closes all devices other than plugin devices and deformable mirrors.
        """

        for device in [
            self.camera,
            self.daq,
            self.remote_focus,
            self.shutter,
            self.zoom,
        ]:
            del device

        for key in list(self.filter_wheel.keys()):
            del self.filter_wheel[key]

        for key in list(self.galvo.keys()):
            del self.galvo[key]

        for key in list(self.laser.keys()):
            del self.laser[key]

        for stage, _ in self.stages_list:
            del stage

    def run_command(self, command: str, *args) -> None:
        """Run command.

        Parameters
        ----------
        command : str
            Command.
        *args : list
            Variable input arguments.
        """
        logger.info(f"Running Command: {command}, {args}")
        if command in self.commands:
            result = self.commands[command][1](*args)
            if result:
                device_name = self.commands[command][0]
                self.output_event_queue.put((device_name, result))
        else:
            logger.debug(f"Unknown Command: {command}")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import threading
import time
import logging

# Third Party Imports
from serial import Serial
from serial import SerialException
from serial import SerialTimeoutException
from serial import EIGHTBITS
from serial import PARITY_NONE
from serial import STOPBITS_ONE
from serial.tools import list_ports

# Local Imports

# Logging setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ASIException(Exception):
    """
    Exception raised when error code from ASI Console is received.

    Attributes:
        - command: error code received from ASI Console
    """

    def __init__(self, code: str) -> None:
        """Initialize the ASIException class
        Parameters
        ----------
        code : str
            Error code received from ASI Console
        """

        #: dict: Dictionary of error codes and their corresponding error messages
        self.error_codes = {
            ":N-1": "Unknown Command (Not Issued in ASI stage(TG-1000/MS2000/MFC2000))",
            ":N-2": "Unrecognized Axis Parameter (valid axes are dependent on the "
            "controller)",
            ":N-3": "Missing parameters (command received requires an axis parameter "
            "such as x=1234)",
            ":N-4": "Parameter Out of Range",
            ":N-5": "Operation failed",
            ":N-6": "Undefined Error (command is incorrect, but the controller does "
            "not know exactly why.",
            ":N-7": "Invalid Card Address",
            ":N-21": "Serial Command halted by the HALT command",
        }
        #: str: Error code received from ASI Console
        self.code = code

        #: str: Error message
        try:
            self.message = self.error_codes[code]
        except KeyError:
            # if the exception is not a standard ASI Error Code:
            self.message = code

        # Gets the proper message based on error code received.
        super().__init__(self.message)
        # Sends message to base exception constructor for python purposes

    def __str__(self) -> str:
        """Overrides base Exception string to be displayed
        in traceback"""
        return f"{self.code} -> {self.message}"


class TigerController:
    """Tiger Controller class"""

    def __init__(self, com_port: str, baud_rate: int, verbose: bool = False) -> None:
        """Initialize the Tiger Controller class


        Parameters
        ----------
        com_port : str
            COM port of the Tiger Controller
        baud_rate : int
            Baud rate of the Tiger Controller
        verbose : bool
            If True, will print out messages to the console

        """
        #: Serial: Serial port object
        self.serial_port = Serial()

        #: str: COM port of the Tiger Controller
        self.com_port = com_port

        #: int: Baud rate of the Tiger Controller
        self.baud_rate = baud_rate

        #: bool: If True, will print out messages to the console
        self.verbose = verbose

        #: list[str]: Default axes sequence of the Tiger Controller
        self.default_axes_sequence = None

        #: list[float]: Maximum speeds of the Tiger Controller
        self._max_speeds = None

        #: threading.Event: Event to indicate if it is safe to write to the serial port
        self.safe_to_write = threading.Event()

        #: threading.Event.set(): Set the safe_to_write event
        self.safe_to_write.set()

        #: float: Last time a command was sent to the Tiger Controller
        self._last_cmd_send_time = time.perf_counter()

    @staticmethod
    def scan_ports() -> list[str]:
        """Scans for available COM ports

        Returns a sorted list of COM ports.

        Returns
        -------
        list[str]
            Sorted list of COM ports
        """
        com_ports = [port.device for port in list_ports.comports()]
        com_ports.sort(key=lambda value: int(value[3:]))
        return com_ports

    def connect_to_serial(
        self,
        rx_size: int = 12800,
        tx_size: int = 12800,
        read_timeout: int = 1,
        write_timeout: int = 1,
    ) -> None:
        """
        Connect to the serial port.

        Parameters
        ----------
        rx_size : int
            Size of the rx buffer
        tx_size : int
            Size of the tx buffer
        read_timeout : int
            Read timeout in seconds
        write_timeout : int
            Write timeout in seconds
        """
        self.serial_port.port = self.com_port
        self.serial_port.baudrate = self.baud_rate
        self.serial_port.parity = PARITY_NONE
        self.serial_port.bytesize = EIGHTBITS
        self.serial_port.stopbits = STOPBITS_ONE
        self.serial_port.xonoff = False
        self.serial_port.rtscts = False
        self.serial_port.dsrdtr = False
        self.serial_port.write_timeout = write_timeout
        self.serial_port.timeout = read_timeout

        # set the size of the rx and tx buffers before calling open
        self.serial_port.set_buffer_size(rx_size, tx_size)

        # try to open the serial port
        try:
            self.serial_port.open()
        except SerialException:
            self.report_to_console(
                f"SerialException: can't connect to {self.com_port} at "
                f"{self.baud_rate}!"
            )

        if self.is_open():
            # clear the rx and tx buffers
            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()
            # report connection status to user
            self.report_to_console("Connected to the serial port.")
            self.report_to_console(
                f"Serial port = {self.com_port} :: Baud rate = {self.baud_rate}"
            )

            #: list[str]: Default axes sequence of the Tiger Controller
            self.default_axes_sequence = self.get_default_motor_axis_sequence()

    def get_default_motor_axis_sequence(self) -> list[str]:
        """Get the default motor axis sequence from the ASI device

        Only returns stage types XYMotor, ZMotor, Theta. Avoids problems associated
        with other tiger controller cards, which could include piezos, filter wheels,
        logic cards, etc.

        Returns
        -------
        list[str]
            Default motor axis sequence
        """
        self.send_command("BU X")
        response = self.read_response()
        lines = response.split("\r")
        motor_axes, axis_types = [], []
        for line in lines:
            if line.startswith("Motor Axes:"):
                motor_axes = line.split("Motor Axes:")[1].split()
            if line.startswith("Axis Types:"):
                axis_types = line.split("Axis Types:")[1].split()

        if len(motor_axes) == 0 or len(axis_types) == 0:
            raise ASIException(":N-5")

        if len(motor_axes) != len(axis_types):
            raise ASIException(":N-5")

        for i in range(len(axis_types) - 1, -1, -1):
            if axis_types[i] not in ["x", "z", "t"]:
                motor_axes.pop(i)
        return motor_axes

    ##### TODO: Modify to accept dictionaries and send a single command for all axes
    def set_feedback_alignment(self, axis: str, aa: float) -> None:
        """Set the stage feedback alignment.

        Adjusts the drive strength by writing to a non-volatile on-board
        potentiometer. Normally done once at the factory (to a very conservative
        value) and never adjusted again. If the AA is off, the stage may be sluggish
        (too low) or it may oscillate, buzz, or sound like it's grinding (too high).
        After changing the AA value the AZERO command should be run until zeroed.

        To optimize stage performance a high AA is desirable, but too high and there
        are big problems. AA can be increased until oscillations occur and then
        decreased by 1 or 2 as described at the page on tuning stages to minimize
        move time.

        Parameters
        ----------
        axis : str
            Stage axis
        aa : float
            Hardware potentiometer value
        """
        self.send_command(f"AA {axis}={aa}\r")
        self.read_response()
        self.send_command(f"AZ {axis}\r")
        self.read_response()

    def set_backlash(self, axis: str, val: float) -> None:
        """Enable/disable stage backlash correction.

        This command sets (or displays) the amount of distance in millimeters of the
        anti-backlash move which absorbs the lash in the axis' gearing at the end of
        commanded moves1). This behind-the-scenes move ensures that the controller
        approaches the final target from the same direction, which improves
        repeatability when using rotary encoders. A value of zero (0) disables the
        anti-backlash algorithm for that axis. The default value depends on motor
        build but is 0.04 for most common 4 TPI leadscrew pitch with rotary encoder,
        0.01 for most common 16 TPI leadscrew pitch, and 0.02 for the x-axis of
        scan-optimized stages. For linear encoders a backlash move is not necessary
        and there is no reason to change the setting from the default value of zero (
        0). Moves with manual input devices (joystick or knobs) do not have any
        anti-backlash move.

        Parameters
        ----------
        axis : str
            Stage axis
        val : float
            Distance of anti-backlash motion [mm]
        """
        self.send_command(f"B {axis}={val:.7f}\r")
        self.read_response()

    def set_finishing_accuracy(self, axis: str, ac: float) -> None:
        """Set the stage finishing accuracy.

        This command sets/displays the Finish Error setting, which controls when the
        motor algorithm routines will turn off. The setting controls the crossover
        position error (in millimeters) between the target and position at which the
        controller will stop attempting to move the stage closer to achieve the ideal
        position=target. This is value also determines the maximum error allowable
        before a move is considered complete. This value is by default set to the
        value of the smallest move step size according to the encoder resolution,
        but many applications do not require such tight landing tolerance.

        Parameters
        ----------
        axis : str
            Stage axis
        ac : float
            Position error [mm]
        """
        self.send_command(f"PC {axis}={ac:.7f}\r")
        self.read_response()

    def set_error(self, axis: str, ac: float) -> None:
        """Set the stage drift error

        This command sets the Drift Error setting. This setting controls the
        crossover position error (in millimeters) between the target and position at
        which the controller considers an axis to be too far out of position when
        resting. When this limit is reached, the controller will re-attempt to move
        the axis back within the Finish Error (PC) limit. The current value for this
        setting can be viewed using the INFO command or by placing a ? after the axis
        name. Entries of zero value, e.g., ERROR X=0 are ignored.

        Parameters
        ----------
        axis : str
            Stage axis
        ac : float
            Position error [mm]
        """
        self.send_command(f"E {axis}={ac:.7f}\r")
        self.read_response()

    ##### END TODO #####

    def disconnect_from_serial(self) -> None:
        """Disconnect from the serial port if it's open."""
        if self.is_open():
            self.serial_port.close()
            self.report_to_console("Disconnected from the serial port.")

    def is_open(self) -> bool:
        """Returns True if the serial port exists and is open.

        Returns
        -------
        bool
            True if the serial port exists and is open
        """
        # short circuits if serial port is None
        return self.serial_port and self.serial_port.is_open

    def report_to_console(self, message: str) -> None:
        """Print message to the output device, usually the console.

        Parameters
        ----------
        message : str
            Message to print to the output device
        """
        # useful if we want to output data to something other than the console
        # (ui element etc)
        if self.verbose:
            print(message)

    def send_command(self, cmd: str) -> None:
        """Send a serial command to the device.

        Parameters
        ----------
        cmd : str
            Serial command to send to the device
        """
        # always reset the buffers before a new command is sent
        self.safe_to_write.wait()
        self.safe_to_write.clear()
        self.serial_port.read_all()
        self.serial_port.reset_input_buffer()
        self.serial_port.reset_output_buffer()

        # send the serial command to the controller
        self.report_to_console(cmd)
        command = bytes(f"{cmd}\r", encoding="ascii")
        try:
            self.serial_port.write(command)
        except SerialTimeoutException as e:
            print(f"Tiger Controller -- SerialTimeoutException: {e}")
            pass

    def read_response(self) -> str:
        """Read a line from the serial response.

        Returns
        -------
        str
            Response from the serial port
        """
        response = self.serial_port.readline()
        self.safe_to_write.set()

        try:
            response = response.decode(encoding="ascii")
        except UnicodeDecodeError:
            """
            Low-Frequency Serial Communication error:
            utf-8' codec can't decode byte 0x8c in position 0:
            invalid start byte
            """
            return ""

        # Remove leading and trailing empty spaces
        response = response.strip()
        self.report_to_console(f"Received Response: {response}")
        if response.startswith(":N"):
            logger.error(f"{str(self)}, Error code received: {response}")
            raise ASIException(response)
        return response  # in case we want to read the response

    def moverel(self, x: int = 0, y: int = 0, z: int = 0) -> None:
        """Move the stage with a relative move on multiple axes.

        Parameters
        ----------
        x : int
            Relative move on the x-axis
        y : int
            Relative move on the y-axis
        z : int
            Relative move on the z-axis
        """
        self.send_command(f"MOVREL X={x} Y={y} Z={z}\r")
        self.read_response()

    def moverel_axis(self, axis: str, distance: float) -> None:
        """Move the stage with a relative move on one axis

        Parameters
        ----------
        axis : str
            Stage axis
        distance : float
            Relative move distance
        """
        self.send_command(f"MOVREL {axis}={round(distance, 6)}\r")
        self.read_response()

    def move(self, pos_dict) -> None:
        """Move the stage with an absolute move on multiple axes

        Parameters
        ----------
        pos_dict : dict
            Dictionary of the form {axis: position}
        """
        pos_str = " ".join(
            [f"{axis}={round(pos, 6)}" for axis, pos in pos_dict.items()]
        )
        self.send_command(f"MOVE {pos_str}\r")
        self.read_response()

    def move_axis(self, axis: str, distance: float) -> None:
        """Move the stage with an absolute move on one axis

        Parameters
        ----------
        axis : str
            Stage axis
        distance : float
            Absolute move distance
        """
        self.send_command(f"MOVE {axis}={round(distance, 6)}\r")
        self.read_response()

    def set_max_speed(self, axis: str, speed: float) -> None:
        """Set the speed on a specific axis. Speed is in mm/s.

        Parameters
        ----------
        axis : str
            Stage axis
        speed : float
            Speed in mm/s
        """
        self.send_command(f"SPEED {axis}={speed}\r")
        self.read_response()

    def get_axis_position(self, axis: str) -> int:
        """Return the position of the stage in ASI units (tenths of microns).

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        int
            Position of the stage in ASI units
        """
        self.send_command(f"WHERE {axis}\r")
        response = self.read_response()
        # try:
        pos = float(response.split(" ")[1])
        # except:
        #     pos = float('Inf')
        return pos

    def get_axis_position_um(self, axis: str) -> float:
        """Return the position of the stage in microns.

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        float
            Position of the stage in microns
        """
        self.send_command(f"WHERE {axis}\r")
        response = self.read_response()
        return float(response.split(" ")[1]) / 10.0

    def get_position(self, axes: list[str]) -> dict:
        """Return current stage position in ASI units.

        If default axes sequence has gotten from the ASI device,
        then it will ask the device all the position in one command,
        else it will ask each axis position one by one.

        WATCH OUT! This will return the positions in the order
        of the underlying hardware no matter what order the axes
        are passed in.

        See https://asiimaging.com/docs/products/serial_commands#commandwhere_w

        Parameters
        ----------
        axes : list[str]
            List of axes to query

        Returns
        -------
        dictionary:
             {axis: position}
        """

        if self.default_axes_sequence:
            cmd = f"WHERE {' '.join(axes)}\r"
            self.send_command(cmd)
            response = self.read_response()

            # return response.split(" ")[1:-1]
            pos = response.split(" ")
            axes_seq = list(
                filter(
                    lambda axis: axis if axis in axes else False,
                    self.default_axes_sequence,
                )
            )
            # return {axis: float(pos[1 + i]) for i, axis in enumerate(axes_seq)}
            axis_dict = {}
            for i, axis in enumerate(axes_seq):
                try:
                    axis_dict[axis] = float(pos[1 + i])
                except (ValueError, IndexError):
                    # Report position failed. Don't crash, we can try again.
                    pass
            return axis_dict
        else:
            result = {}
            for axis in axes:
                result[axis] = self.get_axis_position(axis)
            return result

    # Utility Functions

    def is_axis_busy(self, axis: str) -> bool:
        """Returns True if the axis is busy.

        Parameters
        ----------
        axis : str
            Stage axis
        Returns
        -------
        bool
            True if the axis is busy
        """
        self.send_command(f"RS {axis}?\r")
        res = self.read_response()
        return "B" in res

    def is_device_busy(self) -> bool:
        """Returns True if any axis is busy.

        Returns
        -------
        bool
            True if any axis is busy
        """
        self.send_command("/")
        res = self.read_response()
        return "B" in res

    def wait_for_device(self, timeout: float = 1.75) -> None:
        """Waits for the all motors to stop moving.

        timeout : float
            Timeout in seconds. Default is 1.75 seconds.
        """
        if self.verbose:
            print("Waiting for device...")
        busy = self.is_device_busy()
        waiting_time = 0.0

        while busy:
            waiting_time += 0.001
            if waiting_time >= timeout:
                break
            time.sleep(0.001)
            busy = self.is_device_busy()

        if self.verbose:
            print(f"Waited {waiting_time:.2f} s")

    def stop(self) -> None:
        """Stop all stage movement immediately"""

        self.send_command("HALT")
        self.read_response()
        if self.verbose:
            print("ASI Stages stopped successfully")

    def set_speed(self, speed_dict: dict) -> None:
        """Set speed

        Parameters
        ----------
        speed_dict : dict
            Dictionary of the form {axis: speed}
        """
        axes = " ".join([f"{x}={round(v, 6)}" for x, v in speed_dict.items()])
        self.send_command(f"S {axes}")
        self.read_response()

    def set_speed_as_percent_max(self, pct: float) -> None:
        """Set speed as a percentage of the maximum speed

        Parameters
        ----------
        pct : float
            Percentage of the maximum speed
        """
        if self.default_axes_sequence is None:
            logger.error(
                f"{str(self)}, Default axes sequence is not set. Cannot set speed."
            )
            raise ASIException(
                "Unable to query system for axis sequence. Cannot set speed."
            )
        if self._max_speeds is None:
            # First, set the speed crazy high
            self.send_command(
                f"SPEED {' '.join([f'{ax}=1000' for ax in self.default_axes_sequence])}\r"  # noqa
            )
            self.read_response()

            # Next query the maximum speed
            self.send_command(
                f"SPEED {' '.join([f'{ax}?' for ax in self.default_axes_sequence])}\r"
            )
            res = self.read_response()
            self._max_speeds = [float(x.split("=")[1]) for x in res.split()[1:]]

        # Now set to pct
        self.send_command(
            f"SPEED {' '.join([f'{ax}={pct*speed:.7f}' for ax, speed in zip(self.default_axes_sequence, self._max_speeds)])}\r"  # noqa
        )
        self.read_response()

    def get_speed(self, axis: str) -> float:
        """Get speed

        Parameters
        ----------
        axis : str
            Stage axis
        """
        self.send_command(f"SPEED {axis}?")
        response = self.read_response()
        return float(response.split("=")[1])

    def get_encoder_counts_per_mm(self, axis: str) -> float:
        """Get encoder counts pre mm of axis

        Parameters
        ----------
        axis : str
            Stage axis

        Returns
        -------
        float
            Encoder counts per mm of axis
        """

        self.send_command(f"CNTS {axis}?")
        response = self.read_response()
        return float(response.split("=")[1].split()[0])

    def scanr(
        self,
        start_position_mm: float,
        end_position_mm: float,
        enc_divide: float = 0,
        axis: str = "X",
    ) -> None:
        """Set scanr operation mode.

        Parameters
        ----------
        start_position_mm : float
            Start position in mm
        end_position_mm : float
            End position in mm
        enc_divide : float
            Encoder divide
        axis : str
            Stage axis
        """
        enc_divide_mm = self.get_encoder_counts_per_mm(axis)
        if enc_divide == 0:
            enc_divide = enc_divide_mm
        else:
            enc_divide = int(enc_divide * enc_divide_mm)

        command = (
            f"SCANR "
            f"X={round(start_position_mm, 6)} "
            f"Y={round(end_position_mm, 6)} "
            f"Z={round(enc_divide)}"
        )

        self.send_command(command)
        self.read_response()

    def scanv(
        self,
        start_position_mm: float,
        end_position_mm: float,
        number_of_lines: float,
        overshoot: float = 1.0,
        axis: str = "X",
    ) -> None:
        """Set scanv operation mode.

        Parameters
        ----------
        start_position_mm : float
            Start position in mm
        end_position_mm : float
            End position in mm
        number_of_lines : float
            Number of lines
        overshoot : float
            Overshoot
        axis : str
            Stage axis
        """
        command = (
            f"SCANV "
            f"X={round(start_position_mm, 6)} "
            f"Y={round(end_position_mm, 6)} "
            f"Z={round(number_of_lines, 6)}"
            f"F={round(overshoot, 6)}"
        )

        self.send_command(command)
        self.read_response()

    def start_scan(self, axis: str, is_single_axis_scan: bool = True) -> None:
        """Start scan

        Parameters
        ----------
        axis : str
            Stage axis
        is_single_axis_scan : bool
            If True, will only scan on one axis
        """
        self.send_command("SCAN")
        self.read_response()

    def stop_scan(self) -> None:
        """Stop scan."""
        self.send_command("SCAN P")
        self.read_response()

    def is_moving(self) -> bool:
        """Check to see if the stage is moving.

        Sends the command / which is equivalent to STATUS

        Gets response:
        N - there are no motors running from a serial command
        B - there is a motor running from a serial command

        Returns
        -------
        response: Bool
            True if any axis is moving. False if not.
        """

        # Calculate duration of time since last command.
        time_since_last_cmd = time.perf_counter() - self._last_cmd_send_time

        # Wait 50 milliseconds before pinging controller again.
        sleep_time = 0.050 - time_since_last_cmd
        if sleep_time > 0:
            time.sleep(sleep_time)

        self.send_command("/")
        response = self.read_response().rstrip().rstrip("\r\n")
        if response == "ACK":
            self.send_command("/")
            response = self.read_response().rstrip().rstrip("\r\n")
        if response == "B":
            return True
        elif response == "N":
            return False
        else:
            print("WARNING: WAIT UNTIL DONE RECEIVED NO RESPONSE")
            return False

    # Basic Serial Commands for Filter Wheels

    def send_filter_wheel_command(self, cmd: str) -> None:
        """Send a serial command to the filter wheel.

        Parameters
        ----------
        cmd : str
            Serial command to send to the filter wheel
        """

        # send the serial command to the controller
        self.send_command(f"{cmd}\n")
        # print(f"Sent Command: {command.decode(encoding='ascii')}")

    def select_filter_wheel(self, filter_wheel_number: int = 0) -> None:
        """
        Select the filter wheel, e.g., 0, 1...

        Sets the current filter wheel for subsequent commands. Prompt shows currently
        selected wheel, e.g., 0> is result of FW 0 command. If the selected wheel is
        HOMED and ready to go, the FW command returns the selected wheel as normal. If
        the wheel is not ready for any reason, the response ERR is returned. Example:

        0> FW 1 1 Normal – switch to FW 1
        1> FW 0 ERR FW 0 not ready
        0> Although FW 0 not ready – can still change FW 0 parameters.

        Parameters
        ----------
        filter_wheel_number : int
            The filter wheel number to select.
        """
        self.send_filter_wheel_command(f"FW {filter_wheel_number}")
        self.read_response()

    def move_filter_wheel(self, filter_wheel_position: int = 0) -> None:
        """Move Filter Wheel

        Move to filter position n , where n is a valid filter position.

        Parameters
        ----------
        filter_wheel_position : int
            The position to move the filter wheel to.
        """
        assert filter_wheel_position in range(8)
        self.send_filter_wheel_command(f"MP {filter_wheel_position}")
        self.read_response()

    def move_filter_wheel_to_home(self) -> None:
        """Move the Filter Wheel to Home Position

        Causes current wheel to seek its home position.
        """
        self.send_filter_wheel_command("HO")
        self.read_response()

    def change_filter_wheel_speed(self, speed: int = 0) -> None:
        """Change the Filter Wheel Speed

        Selects a consistent set of preset acceleration and speed parameters.
        Supported in version 2.4 and later.

        Parameters
        ----------
        speed : int
            Speed of the filter wheel. The speed can be set to one of the following values:
        - 0: Default - directly set and saved AU, AD, and VR parameters are used.
        - 1: Slowest and smoothest switching speed.
        - 2 to 8: Intermediate switching speeds.
        - 9: Fastest but least reliable switching speed.
        """
        self.send_filter_wheel_command(f"SV {speed}")
        self.read_response()

    def halt_filter_wheel(self) -> None:
        """Halt filter wheel"""
        self.send_filter_wheel_command("HA")
        self.read_response()

    def move_dichroic(self, dichroic_id: str, dichroic_position: int = 0) -> None:
        """Move Dichroic Slider.

        Move to dichroic position n , where n is a valid filter position.
        The Motorized C60 Cube Slider from ASI has 4 positions.

        Parameters
        ----------
        dichroic_id : str
            The ID of the dichroic in the Tiger Controller. e.g., "T"
        dichroic_position : int
            The position to move the dichroic to.
        """
        self.send_filter_wheel_command(f"MOVE {dichroic_id}={dichroic_position}")
        self.read_response()

    def square_wave(self, on_time: int, delay_time: int) -> None:
        """Square wave modulation.

        For testing only. Can be used to implement precise laser control.

        Parameters
        ----------
        on_time : int
            On time in quarter milliseconds
        delay_time : int
            Delay time in quarter milliseconds
        """
        commands = [
            "CCA X=0",
            "M E=2",
            "CCA Y=15",
            f"CCA Z={delay_time}",
            "CCB X=68 Y=192 Z=0",
            "M E=3",
            "CCA Y=5",
            "CCB X=1 Y=66",
            "M E=4",
            "CCA Y=14",
            f"CCA Z={on_time}",
            "CCB X=3 Y=192 Z=0",
            "M E=33",
            "CCA Z=1",
            "M E=34",
            "CCA Z=2",
            "M E=35",
            "CCA Z=3",
            "M E=36",
            "CCA Z=4",
            "M E=1",
            "CCA Z=1",
        ]
        for command in commands:
            # Send data
            self.send_command(f"{command}\r")
            self.read_response()

    def logic_card_on(self, axis: str):
        """Turn on the logic card output

        Parameters
        ----------
        axis : str
            The axis of the logic card output
        """
        axis = int(axis) + 32
        self.send_command(f"6 M E = {axis}\r")
        self.read_response()
        self.send_command(f"6 CCA Z=64\r")
        self.read_response()

    def logic_card_off(self, axis: str):
        """Turn off the logic card output

        Parameters
        ----------
        axis : str
            The axis of the logic card output
        """
        axis = int(axis) + 32
        self.send_command(f"6 M E = {axis}\r")
        self.read_response()
        self.send_command(f"6 CCA Z=0\r")
        self.read_response()

    def logic_cell_on(self, axis : str):
        """Turn on internal logic cell

        Parameters
        ----------
        axis : str
            The axis of the internal logic cell
        """
        self.send_command(f'6 M E = {axis}\r')
        self.read_response()
        self.send_command(f'6 CCA Z=1\r')
        self.read_response()

    def logic_cell_off(self, axis :str):
        """Turn off internal logic cell

        Parameters
        ----------
        axis : str
            The axis of the internal logic cell
        """
        self.send_command(f'6 M E = {axis}\r')
        self.read_response()
        self.send_command(f'6 CCA Z=0\r')
        self.read_response()

    def SA_waveform(
        self, axis: str, waveform: int = 0, amplitude: int = 1000, offset: int = 500, period: int = 10 
    ) -> None:
        """Programs the analog waveforms using SAA, SAO, SAP, and SAF
        Default waveform is a sawtooth waveform with an amplitude of 1V, an offset of 0.5V and period of 10 ms

        Parameters
        ----------
        axis: str
            Tiger Controller axis
        waveform: int
            Type of waveform pattern according to https://asiimaging.com/docs/commands/sap
        amplitude: int
            amplitude of the waveform in mV
        offset: int
            sets the center position of the waveform in mV
        period: int
            sets the period of the waveform in ms
        """
        print(f"Period (ms): {period}")
        # takes amplitude and offset from navigate and modifies them to how the TG-1000 takes them
        if (waveform % 128 == 3):
            offset = .5*(offset+amplitude)

        amplitude = amplitude*2

        print("***", waveform, amplitude, axis, offset, period)
        # TODO: 3 is the address of the GALVO DAC. May need to make this configurable.
        self.send_command(f"3 SAP {axis}={round(waveform)}")
        self.read_response()
        self.send_command(f"3 SAA {axis}={round(amplitude)}")
        self.read_response()
        self.send_command(f"3 SAO {axis}={round(offset)}")
        self.read_response()
        self.send_command(f"3 SAF {axis}={round(period)}")
        self.read_response()

    def SAM(self, axis: str, mode: int) -> None:
        """Sets the single-axis mode according to the integer code.

        0: stops waveforms if they are running
        1: starts generating the waveform pattern
        2: waveform only runs for one cycle, then waits for another trigger
        3: starts generating the waveform pattern, restarts the other waveform on the same card
        4: starts generating the waveform, free running after the trigger

        Parameters
        ----------
        axis: str
            Laser axis
        mode: int
            Integer code.
        """
        self.send_command(f"3 SAM {axis}={mode}")
        self.read_response()

    def setup_control_loop(
        self,
        delays: list[float],
        camera_delay: float,
        remote_focus_delay: float,
        sweep_time: float,
        analog_outputs: dict,
    ) -> None:
        """
        Sets up the control loop for triggering the remote focus, Galvo/s, and Camera
        Trigger

        If/Else statements: Used to handle either single or multiple galvo cases

        Parameters
        ----------
        delays: list[float]
            One or two floats that are the delays for the galvos triggering
        camera_delay: float
            Delay in ms for the camera to turn on relative to the master trigger
        remote_focus_delay: float
            Delay in ms for the remote focus to turn on relative to the master trigger
        sweep_time: float
            THe time in ms between each iteration of the master trigger
        analog_outputs: dict
            Dictionary that includes the device and the output axis on the DAC4 card,
            as specified in config
        """
        # TODO: Investigate if these axis outputs are shared amongst units.
        # Reference values for ttls that correspond to outputs A-C
        ttls = {"A": 42, "B": 44, "C": 46}

        start_delay = int(delays[0] * 4)  # Unit conversion from ms to 1/4 ms

        # If/Else to handle either 1 or two galvos
        if len(delays) > 1:
            galvo2_delay = int((delays[0] - delays[1]) * 4)
            galvo1_axis = analog_outputs["galvo 0"]
            galvo2_axis = analog_outputs["galvo 1"]
        elif len(delays) == 1:
            galvo1_axis = analog_outputs["galvo 0"]
            galvo2_delay = 0
        else:
            galvo2_delay = 0
        remote_focus_axis = analog_outputs["remote_focus"]

        sweep_time = (
            int(sweep_time * 4) - 2
        )  # Hardcoded -2 to account for delays within controller

        # Dynamic delay processing. Instead of having each delay handled separately,
        # to save some cell space delays are programmed based on some simple
        # calculation logic. Cell 6 is the direct output for the logical loop,
        # and cell 12 is the post-delay output. So whichever delay is longer
        # (accounting for hardware delays), will get input from 12. Additionally,
        # delay is calculated as a shifted difference between the higher and lower delay
        if remote_focus_delay > camera_delay + 2:
            camera_output = 6
            remote_focus_output = 12
            start_delay += int((camera_delay + 2) * 4)
            difference_delay = int((remote_focus_delay - (camera_delay + 2)) * 4)
        elif remote_focus_delay < camera_delay + 2:
            camera_output = 12
            remote_focus_output = 6
            start_delay += int(remote_focus_delay * 4)
            difference_delay = int(((camera_delay + 2) - remote_focus_delay) * 4)
        else:
            camera_output = 6
            remote_focus_output = 6
            start_delay += int((camera_delay + 2) * 4)
            difference_delay = 0

        commands = [
            # Resets programmable logic cell configurations to constant cells
            # This handles configuring cell 1 and cell 8
            "6 CCA X=0",
            # Cell 2, a one-shot triggered by the rising edge of Cell 1
            "6 m e = 2",
            "6 cca y = 8",
            "6 cca z = 10",
            "6 ccb x = 1",
            "6 ccb y = 192",
            # Cell 3, delay cell that is used to sync up the loop after Galvo initialization
            # Delay time is based on the start delay variable (in 1/4 ms)
            "6 m e = 3",
            "6 cca y = 9",
            f"6 cca z = {start_delay}",
            "6 ccb x = 1",
            "6 ccb y = 192",
            # Cell 4, JK flop used for toggling on and off state of the loop
            # Cells 3 and 8 serve as the inputs of this cell
            "6 m e = 4",
            "6 cca y = 13",
            "6 ccb x = 3",
            "6 ccb y = 8",
            "6 ccb z = 192",
            # Cell 5, AND cell used to check if the loop is still operating
            # Cell inputs are the output of cell 4 and the inverse of cell 7 (64+7)
            "6 m e = 5",
            "6 cca y = 5",
            "6 ccb x = 4",
            "6 ccb y = 71",
            # Cell 6, a one-shot triggered by the rising edge of Cell 5
            "6 m e = 6",
            "6 cca y = 8",
            "6 cca z = 10",
            "6 ccb x = 5",
            "6 ccb y = 192",
            # Cell 7, delay cell that waits for the sweep time until retriggering.
            # Used for the main loop. Timing is dependent on the sweep_time variable
            "6 m e = 7",
            "6 cca y = 9",
            f"6 cca z= {sweep_time}",
            "6 ccb x = 6",
            "6 ccb y = 192",
            # Cell 9, delay used to sync the phase of the Galvos
            # This is because the Tiger Controller can not arbitrarily start waveforms
            "6 m e = 9",
            "6 cca y = 9",
            f"6 cca z = {galvo2_delay}",
            "6 ccb x = 2",
            "6 ccb y = 192",
            # Cell 10, a one-shot triggered by the rising edge of Cell 9
            "6 m e = 10",
            "6 cca y = 8",
            "6 cca z = 10",
            "6 ccb x = 9",
            "6 ccb y = 192",
            # Cell 11, delay cell used to sync up the camera trigger and remote focus
            # output based on configuration
            "6 m e = 11",
            "6 cca y = 9",
            f"6 cca z = {difference_delay}",
            "6 ccb x = 6",
            "6 ccb y = 192",
            # Cell 12, a one-shot triggered by the rising edge of Cell 11
            "6 m e = 12",
            "6 cca y = 8",
            "6 cca z = 10",
            "6 ccb x = 11",
            "6 ccb y = 192",
            # Routes the output of the remote focus trigger to the TTL output from the
            # PLC
            f"6 m e = {ttls[remote_focus_axis]+1}",
            "6 cca y = 1",
            f"6 cca z = {remote_focus_output}",
            # Routes the TTL output from the previous section to a different TTL to
            # actually trigger the waveform
            f"6 m e = {ttls[remote_focus_axis]}",
            "6 cca y = 1",
            f"6 cca z = {ttls[remote_focus_axis]+1}",
            # Sets the camera signal output to the first physical PLC output
            "6 m e = 33",
            f"6 cca z = {camera_output}",
        ]
        # Creates object to hold galvo commands
        galvo_commands = []
        # Single Galvo case, just sets up the first Galvo
        if len(delays) == 1:
            galvo_commands = [
                # Sets the output of Cell 2 as the input to the TTL corresponding to
                # the first Galvo pair
                f"6 m e = {ttls[galvo1_axis]+1}",
                "6 cca y = 1",
                "6 cca z = 2",
                # Sets the output of the first Galvo TTL to the Galvo TTL that
                # actually triggers it
                f"6 m e = {ttls[galvo1_axis]}",
                "6 cca y = 1",
                f"6 cca z = {ttls[galvo1_axis]+1}",
            ]
        # Multiple Galvo case, has the first set of commands and the commands for the
        # second Galvo
        elif len(delays) > 1:
            galvo_commands = [
                # Sets the output of Cell 2 as the input to the TTL corresponding to
                # the first Galvo pair
                f"6 m e = {ttls[galvo1_axis]+1}",
                "6 cca y = 1",
                "6 cca z = 2",
                # Sets the output of the first Galvo TTL to the Galvo TTL that
                # actually triggers it
                f"6 m e = {ttls[galvo1_axis]}",
                "6 cca y = 1",
                f"6 cca z = {ttls[galvo1_axis]+1}",
                # Sets the output of Cell 10 as the input to the TTL corresponding to
                # the second Galvo pair
                f"6 m e = {ttls[galvo2_axis]+1}",
                "6 cca y = 1",
                "6 cca z = 10",
                # Sets the output of the second Galvo TTL to the Galvo TTL that
                # actually triggers it
                f"6 m e = {ttls[galvo2_axis]}",
                "6 cca y = 1",
                f"6 cca z = {ttls[galvo2_axis]+1}",
            ]

        # TODO: Evaluate if commands can be combined to reduce the number of commands
        #  sent to reduce latency
        # Runs the main setup commands, followed by the Galvo specific commands
        for command in commands:
            self.send_command(f"{command}\r")
            self.read_response()
        for command in galvo_commands:
            self.send_command(f"{command}\r")
            self.read_response()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time
import io

# Third Party Imports
import serial

# Local Imports
from navigate.model.devices.filter_wheel.base import FilterWheelBase
from navigate.model.devices.device_types import SerialDevice

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class LUDLFilterWheel(FilterWheelBase, SerialDevice):

    """LUDLFilterWheel - Class for controlling LUDL Electronic Products Filter Wheels

    Testing using MAC6000 controller over RS-232. USB or Ethernet not tested.
    Single filter wheel only, but MAC6000 can handle max 2. Would need to modify code.

    Note
    ----
        Code for class adapted from Dr. Fabian Voigt's mesoSPIM code:
        https://github.com/ffvoigt/mesoSPIM-control/blob/master/mesoSPIM/src/mesoSPIM_FilterWheel.py

    """

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the LUDLFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : Any
            The communication instance with the device.
        configuration : Dict[str, Any]
            Global configuration dictionary.
        device_id : int
            The ID of the device. Default is 0.
        """

        super().__init__(microscope_name, device_connection, configuration, device_id)

        #: Any: Device connection object.
        self.serial = device_connection

        #: io.TextIOWrapper: Text I/O wrapper for the serial port.
        self.sio = io.TextIOWrapper(io.BufferedRWPair(self.serial, self.serial))

        #: float: Delay for filter wheel to change positions.
        self.wait_until_done_delay = self.device_config["filter_wheel_delay"]

    def __str__(self):
        """String representation of the class."""
        return "LUDLFilterWheel"

    @classmethod
    def connect(cls, comport, baudrate, timeout=0.25):
        """Build LUDLFilterWheel Serial Port connection

        Attributes
        ----------
        comport : str
            Comport for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, e.g., 9600.
        timeout : float
            Timeout for communicating with the filter wheel, e.g., 0.25.

        Returns
        -------
        serial.Serial
            Serial port connection to the filter wheel.

        Raises
        ------
        UserWarning
            Could not communicate with LUDL MAC6000 via COMPORT.
        """
        logging.debug(f"LUDLFilterWheel - Opening Serial Port {comport}")
        try:
            return serial.Serial(
                comport,
                baudrate,
                parity=serial.PARITY_NONE,
                timeout=timeout,
                xonxoff=False,
                stopbits=serial.STOPBITS_TWO,
            )
        except serial.SerialException:
            logger.error("LUDLFilterWheel - Could not establish Serial Port Connection")
            raise UserWarning(
                "Could not communicate with LUDL MAC6000 via COMPORT", comport
            )

    def set_filter(self, filter_name, wait_until_done=True):
        """Set the filter wheel to a specific filter position.

        Parameters
        ----------
        filter_name : str
            Name of the filter position.
        wait_until_done : bool
            Wait until the filter wheel has changed positions.
        """

        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # single wheel code only...

            # Set the new position by cross-referencing the filter dict
            self.wheel_position = self.filter_dictionary[filter_name]

            # Send the command to go to the new position
            self.sio.write(f"Rotat S M {self.wheel_position}\n")
            self.sio.flush()

            if wait_until_done:
                time.sleep(self.wait_until_done_delay)

    def close(self):
        """Close the LUDLFilterWheel serial port.

        Sets the filter wheel to the Empty-Alignment position and then closes the port.
        """
        logger.debug("LUDLFilterWheel - Closing the Filter Wheel Serial Port")
        self.set_filter(list(self.filter_dictionary.keys())[0])
        self.serial.close()

    def __del__(self):
        """Destructor for the LUDLFilterWheel class."""
        if self.serial.is_open:
            self.close()



"""
Sub-controllers for navigate.
"""

from .stages import StageController  # noqa
from .acquire_bar import AcquireBarController  # noqa
from .channels_tab import ChannelsTabController  # noqa
from .camera_view import CameraViewController, MIPViewController  # noqa
from .camera_settings import CameraSettingController  # noqa
from .waveform_tab import WaveformTabController  # noqa
from .waveform_popup import WaveformPopupController  # noqa
from .autofocus import AutofocusPopupController  # noqa
from .features_popup import FeaturePopupController  # noqa
from .feature_advanced_setting import FeatureAdvancedSettingController  # noqa
from .keystrokes import KeystrokeController  # noqa
from .multiposition import MultiPositionController  # noqa
from .ilastik import IlastikPopupController  # noqa
from .camera_map import CameraMapSettingPopupController  # noqa
from .microscope_popup import MicroscopePopupController  # noqa
from .adaptive_optics import AdaptiveOpticsPopupController  # noqa
from .histogram import HistogramController  # noqa

# from .uninstall_plugin_controller import UninstallPluginController  # noqa
from .plugins import PluginsController, UninstallPluginController  # noqa
from .menus import MenuController  # noqa



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard library imports
import importlib
from importlib.metadata import entry_points
import importlib.resources
import pkgutil
import os
import inspect
import logging
from typing import Optional, Any

# Local application imports
from navigate.tools.file_functions import load_yaml_file
from navigate.tools.common_functions import load_module_from_file
from navigate.model.features import feature_related_functions

# Logger setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def register_features(module) -> None:
    """Register features

    Parameters
    ----------
    module : module
        A python module contains features
    """
    for c in dir(module):
        if inspect.isclass(getattr(module, c)):
            setattr(feature_related_functions, c, getattr(module, c))


class PluginPackageManager:
    """Plugin package manager"""

    @staticmethod
    def get_plugins() -> dict:
        """Get plugins

        Returns
        -------
        plugins : dict
            plugins dict
        """
        plugins = {}
        for entry_point in entry_points().get("navigate.plugins", []):
            plugin_package_name = entry_point.module.split(".")[0]
            if plugin_package_name in plugins:
                error_statement = (
                    f"Warning: plugin {plugin_package_name} exists and cannot be "
                    f"loaded more than once. Please double-check the installed plugins."
                )
                logger.debug(error_statement)
                print(error_statement)
                continue
            plugins[plugin_package_name] = plugin_package_name
        return plugins

    @staticmethod
    def load_config(package_name: str) -> dict:
        """Load plugin_config.yml

        Parameters
        ----------
        package_name : str
            package name

        Returns
        -------
        plugin_config : dict
            plugin configuration
        """
        package_path = importlib.resources.files(package_name)
        plugin_config = load_yaml_file(os.path.join(package_path, "plugin_config.yml"))
        return plugin_config

    @staticmethod
    def load_controller(package_name: str, controller_name: str) -> Optional[Any]:
        """Load controller

        Parameters
        ----------
        package_name : str
            package name
        controller_name : str
            controller name

        Returns
        -------
        controller_class : Optional[Any]
            The dynamically loaded controller class, or `None` if the controller
            module or class cannot be found. The return type can be:

            - `type` : If the controller class is found, the class type is returned.
            - `None` : If the module or class cannot be found, `None` is returned.
        """
        controller_file_name = "_".join(controller_name.lower().split()) + "_controller"
        controller_class_name = "".join(controller_name.title().split()) + "Controller"
        try:
            controller_module = importlib.import_module(
                f"{package_name}.controller.{controller_file_name}"
            )
            return getattr(controller_module, controller_class_name)
        except (ImportError, AttributeError):
            logger.debug("Plugin controller not found.")
            return None

    @staticmethod
    def load_view(package_name: str, frame_name: str) -> Optional[Any]:
        """Load view

        Parameters
        ----------
        package_name : str
            package name
        frame_name : str
            frame name

        Returns
        -------
        view_class : Optional[Any]
            The dynamically loaded view class, or `None` if the view module or
            class cannot be found. The return type can be:

            - `type` : If the view class is found, the class type is returned.
            - `None` : If the module or class cannot be found, `None` is returned.
        """
        frame_file_name = "_".join(frame_name.lower().split()) + "_frame"
        frame_class_name = "".join(frame_name.title().split()) + "Frame"
        try:
            view_module = importlib.import_module(
                f"{package_name}.view.{frame_file_name}"
            )
            return getattr(view_module, frame_class_name)
        except (ImportError, AttributeError):
            logger.debug("Plugin view not found.")
            return None

    @staticmethod
    def load_feature_lists(package_name, register_func) -> None:
        """Load feature lists

        Parameters
        ----------
        package_name : str
            package name
        register_func : func
            the function to handle feature lists
        """
        try:
            module = importlib.import_module(f"{package_name}.feature_list")
            register_func(
                importlib.resources.files(package_name).joinpath("feature_list.py"),
                module,
            )
        except (ImportError, AttributeError):
            logger.debug("Plugin feature list not found.")
            pass

    @staticmethod
    def load_features(package_name):
        """Load features

        Parameters
        ----------
        package_name : str
            package name
        """
        for _, module_name, is_pkg in pkgutil.iter_modules(
            [importlib.resources.files(package_name).joinpath("model/features")]
        ):
            if not is_pkg:
                full_module_name = f"{package_name}.model.features.{module_name}"
                try:
                    module = importlib.import_module(full_module_name)
                except (ImportError, AttributeError):
                    logger.debug("Plugin feature not found.")
                    continue
                register_features(module)

    @staticmethod
    def load_acquisition_modes(package_name, acquisition_modes, register_func):
        """Load acquisition modes

        Parameters
        ----------
        package_name : str
            package name
        acquisition_modes : []
            list of acquisition mode configurations
        register_func : func
            the function to register acquisition modes
        """
        for acquisition_mode_config in acquisition_modes:
            acquisition_file = acquisition_mode_config["file_name"][:-3]
            try:
                module = importlib.import_module(f"{package_name}.{acquisition_file}")
            except (ImportError, AttributeError):
                logger.debug("Plugin acquisition mode not found.")
                continue
            if module:
                register_func(acquisition_mode_config["name"], module)

    @staticmethod
    def load_devices(package_name, register_func):
        """Load devices

        Parameters
        ----------
        package_name : str
            package name
        register_func : func
            the function to register devices
        """
        for _, module_name, is_pkg in pkgutil.iter_modules(
            [importlib.resources.files(package_name).joinpath("model/devices")]
        ):
            if is_pkg:
                full_module_name = (
                    f"{package_name}.model."
                    f"devices.{module_name}.device_startup_functions"
                )
                try:
                    module = importlib.import_module(full_module_name)
                except (ImportError, AttributeError) as e:
                    logger.debug(f"Plugin device not found: {e}")
                    continue
                register_func(module_name, module)


class PluginFileManager:
    """Plugin file manager"""

    def __init__(self, plugins_path: str, plugins_config_path: str) -> None:
        """Initialize PluginFileManager

        Parameters
        ----------
        plugins_path : str
            plugins path
        plugins_config_path : str
            plugins config path
        """

        #: str: plugins path
        self.plugins_path = plugins_path

        #: str: plugins config path
        self.plugins_config_path = plugins_config_path

    def get_plugins(self):
        """Get plugins

        Returns
        -------
        plugins : dict
            plugins dict
        """
        plugins = {}

        folder_names = os.listdir(self.plugins_path)
        full_path_names = dict(
            [(f, os.path.join(self.plugins_path, f)) for f in folder_names]
        )
        # add plugins from plugins_config.yml
        plugins_config = load_yaml_file(self.plugins_config_path)
        if plugins_config:
            full_path_names.update(plugins_config)

        for plugin_name, plugin_path in full_path_names.items():
            if (
                plugin_path
                and os.path.exists(plugin_path)
                and os.path.isdir(plugin_path)
            ):
                plugins[plugin_name] = plugin_path
                logger.info(f"Plugin {plugin_name} found.")

        return plugins

    @staticmethod
    def load_config(plugin_path):
        """Load plugin_config.yml

        Parameters
        ----------
        plugin_path : str
            plugin path

        Returns
        -------
        plugin_config : dict
            plugin configuration
        """
        plugin_config = load_yaml_file(os.path.join(plugin_path, "plugin_config.yml"))
        return plugin_config

    @staticmethod
    def load_controller(plugin_path, controller_name):
        """Load controller

        Parameters
        ----------
        plugin_path : str
            plugin path
        controller_name : str
            controller name

        Returns
        -------
        controller_class : class
            controller class
        """
        controller_file_name = (
            "_".join(controller_name.lower().split()) + "_controller.py"
        )
        controller_class_name = "".join(controller_name.title().split()) + "Controller"
        controller_file_path = os.path.join(
            plugin_path, "controller", controller_file_name
        )
        if os.path.exists(controller_file_path) and os.path.isfile(
            controller_file_path
        ):
            module = load_module_from_file(controller_class_name, controller_file_path)
            return getattr(module, controller_class_name)
        return None

    @staticmethod
    def load_view(plugin_path, frame_name):
        """Load view

        Parameters
        ----------
        plugin_path : str
            plugin path
        frame_name : str
            frame name

        Returns
        -------
        frame_class : class
            tkinter frame class
        """
        frame_file_name = "_".join(frame_name.lower().split()) + "_frame.py"
        frame_class_name = "".join(frame_name.title().split()) + "Frame"
        frame_file_path = os.path.join(plugin_path, "view", frame_file_name)
        if os.path.exists(frame_file_path) and os.path.isfile(frame_file_path):
            module = load_module_from_file(frame_class_name, frame_file_path)
            return getattr(module, frame_class_name)
        return None

    @staticmethod
    def load_feature_lists(plugin_path, register_func):
        """Load feature lists

        Parameters
        ----------
        plugin_path : str
            plugin path
        register_func : func
            the function to handle feature lists
        """
        plugin_feature_list = os.path.join(plugin_path, "feature_list.py")
        if os.path.exists(plugin_feature_list):
            module = load_module_from_file("feature_list_temp", plugin_feature_list)
            register_func(plugin_feature_list, module)

    @staticmethod
    def load_features(plugin_path):
        """Load features

        Parameters
        ----------
        plugin_path : str
            plugin path
        """
        features_dir = os.path.join(plugin_path, "model", "features")
        features = []
        if os.path.exists(features_dir):
            features = os.listdir(features_dir)
        for feature in features:
            feature_file = os.path.join(features_dir, feature)
            if os.path.isfile(feature_file):
                module = load_module_from_file(feature, feature_file)
                register_features(module)

    @staticmethod
    def load_acquisition_modes(plugin_path, acquisition_modes, register_func):
        """Load acquisition modes

        Parameters
        ----------
        plugin_path : str
            plugin path
        acquisition_modes : []
            list of acquisition mode configurations
        register_func : func
            the function to register acquisition modes
        """
        for acquisition_mode_config in acquisition_modes:
            acquisition_file = acquisition_mode_config["file_name"]
            full_path_name = os.path.join(plugin_path, acquisition_file)
            if os.path.exists(full_path_name):
                module = load_module_from_file(acquisition_file[:-3], full_path_name)
                if module:
                    register_func(acquisition_mode_config["name"], module)

    @staticmethod
    def load_devices(plugin_path, register_func):
        """Load devices

        Parameters
        ----------
        plugin_path : str
            plugin path
        register_func : func
            the function to register devices
        """
        device_dir = os.path.join(plugin_path, "model", "devices")
        if os.path.exists(device_dir) and os.path.isdir(device_dir):
            devices = os.listdir(device_dir)
            for device in devices:
                device_path = os.path.join(device_dir, device)
                if not os.path.isdir(device_path):
                    continue
                try:
                    module = load_module_from_file(
                        "device_module",
                        os.path.join(device_path, "device_startup_functions.py"),
                    )
                except FileNotFoundError:
                    logger.debug(f"Plugin device {device} not found.")
                    continue
                register_func(device, module)



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
from typing import Any, Dict

# Third Party Imports

# Local Imports
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class LaserBase:
    """Laser Base Class"""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        device_id: int,
    ) -> None:
        """Initialize Laser Base Class

        Parameters
        ----------
        microscope_name : str
            Name of the microscope
        device_connection : Any
            Communication instance with the device.
        configuration : Dict[str, Any]
            Configuration dictionary
        device_id : int
            Laser ID
        """
        #: Any: Communication instance with the device
        self.device_connection = device_connection

        #: dict: Configuration dictionary
        self.configuration = configuration

        #: str: Name of the microscope
        self.microscope_name = microscope_name

        #: int: Laser ID
        self.device_config = configuration["configuration"]["microscopes"][
            microscope_name
        ]["laser"][device_id]

    def __str__(self) -> str:
        """Return string representation of the class"""
        return "LaserBase"

    def set_power(self, laser_intensity: int) -> None:
        """Set laser power

        Parameters
        ----------
        laser_intensity : int
            Laser intensity
        """
        pass

    def turn_on(self) -> None:
        """Turn on the laser"""
        pass

    def turn_off(self) -> None:
        """Turn off the laser"""
        pass

    def close(self) -> None:
        """
        Close the laser before exit.
        """
        pass

    def initialize_laser(self) -> None:
        """Initialize lasers."""
        pass



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import traceback
from typing import Any, Dict

# Local Imports
from navigate.model.devices.shutter.base import ShutterBase
from navigate.model.devices.device_types import SerialDevice
from navigate.tools.decorators import log_initialization
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASIShutter(ShutterBase, SerialDevice):
    """ShutterTTL Class

    Triggering for shutters delivered from the TigerController.
    Each output keeps their last digital state for as long the device is not
    powered down.
    """

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        address=None,
    ) -> None:
        """Initialize the ASIShutter.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: str: Address of the shutter, if applicable.
        self.address = address

        #: TigerController: ASI Tiger Controller object.
        self.shutter = device_connection

        #: str: Axis of the shutter.
        self.axis = configuration["configuration"]["microscopes"][microscope_name][
            "shutter"
        ]["hardware"]["axis"]

        #: str: Port for communicating with the shutter.
        self.port = configuration["configuration"]["microscopes"][microscope_name][
            "shutter"
        ]["hardware"]["port"]

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASILaser Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the shutter, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the shutter, default is 115200.
        timeout : float
            Timeout for communicating with the shutter, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI shutter connection failed.")
            raise Exception("ASI shutter connection failed.")
        return tiger_controller

    def __del__(self) -> None:
        """Disconnect from the serial port.

        Raises
        -------
        Exception
            If there is an error during the cleanup process.
        """
        try:
            if self.shutter:
                self.shutter.disconnect_from_serial()
                logger.debug("TigerController disconnected successfully.")
        except Exception as e:
            logger.exception(f"Error during cleanup: {traceback.format_exc()}")

    def open_shutter(self) -> None:
        """Opens the shutter.

        Raises
        -------
        Exception
            If there is an error while trying to open the shutter.
        """
        try:
            self.shutter.logic_card_on(self.axis)
            logger.debug("ASIShutter opened")
        except Exception as e:
            logger.exception(f"Shutter not open: {traceback.format_exc()}")

    def close_shutter(self) -> None:
        """Closes the shutter.

        Raises
        -------
        Exception
            If there is an error while trying to close the shutter.
        """
        try:
            self.shutter.logic_card_off(self.axis)
            logger.debug("ASIShutter closed")
        except Exception as e:
            logger.exception(f"Shutter did not close: {traceback.format_exc()}")

    @property
    def state(self) -> bool:
        """Get the state of the shutter.

        Returns
        -------
        bool:
            True if the shutter is open, False if it is closed.
        """
        return self.shutter.get_axis_position(self.axis)
